#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#



from asn1crypto import core
import enum
import os
import io

from responder3.core.commons import *
from responder3.core.asyncio_helpers import *
from responder3.crypto.pure.AES.AES import Counter, AESModeOfOperationCTR

import hashlib
import struct
import rsa
from rsa import pkcs1
import hmac

MAX_PACKET_LENGTH = 35000
zero_byte = b'\x00'


class SSHPrivKey:
	name = 'ssh-rsa'
	def __init__(self, privkey = None):
		self.privkey = privkey
	
	@staticmethod
	def load_privkey_from_string(pk_str):
		"""
		Loads a PEM encoded RSA private key
		Format must be the same as /etc/ssh/ssh_host_rsa_key
		(openssl RSA private key)
		"""
		pk = SSHPrivKey()
		pk.privkey = rsa.PrivateKey.load_pkcs1(pk_str,'PEM')
		return pk
		
	@staticmethod
	def load_privkey_from_file(filename):
		with open(filename, 'rb') as f:
			return SSHPrivKey.load_privkey_from_string(f.read())
		
		
	def sign(self, data):
		"""
		hashes the data with sha1 and signs the result using pkcs1.5 and SHA-1
		returns bytes
		data must be bytes!
		"""
		return pkcs1.sign(data, self.privkey, 'SHA-1')


class SSHstring:
	def __init__(self):
		pass
		
	@staticmethod
	def from_buff(buff):
		length = int.from_bytes(buff.read(4), byteorder = 'big', signed = False)
		if length == 0:
			return None
		data = buff.read(length)
		return data.decode()
		
	@staticmethod
	def to_bytes(s):
		if not s:
			return b''
		
		data = s.encode()
		length = len(data).to_bytes(4, byteorder="big", signed = False)
		return length + data
		
class SSHBytes:
	@staticmethod
	def from_buff(buff):
		length = int.from_bytes(buff.read(4), byteorder = 'big', signed = False)
		if length == 0:
			return None
		return buff.read(length)
		
	@staticmethod
	def to_bytes(s):
		if not s:
			data = b''
			return len(data).to_bytes(4, byteorder="big", signed = False)
		
		length = len(s).to_bytes(4, byteorder="big", signed = False)
		return length + s
		
class MPInt:
	"""
	THIS CLASS IS ONLY FOR POSITIVE INTEGERS!!!
	"""
	def __init__(self):
		pass
		
	@staticmethod
	def from_buff(buff):
		length = int.from_bytes(buff.read(4), byteorder = 'big', signed = False)
		if length == 0:
			return None
		contents = buff.read()
		buff = io.BytesIO(contents)
		data = buff.read(length)
		if data[0] == 0:
			data = data[1:]
		return MPInt.bytes2int(data)
		
	@staticmethod
	def to_bytes(integer):
		if integer is None:
			return b'\x00' * 4
		bdata = MPInt.int2bytes(integer)
		if bdata[0] > 128:
			bdata = b'\x00' + bdata
		return len(bdata).to_bytes(4, byteorder="big", signed = False) + bdata
		
	@staticmethod
	def int2bytes_padded(integer):
		bdata = MPInt.int2bytes(integer)
		if bdata[0] > 128:
			bdata = b'\x00' + bdata
		return bdata
		
	@staticmethod
	def bytes2int(s):
		return int.from_bytes(s, 'big', signed = False)
		
	@staticmethod
	def int2bytes(integer):
		"""
		Only for positive integers!
		"""
		hd = hex(integer)[2:]
		if len(hd) %2:
			hd = '0' + hd
		
		return bytes.fromhex(hd)
		#return integer.to_bytes((integer.bit_length() + 7) // 8, byteorder = 'big', signed = False)
	
		

class NameList:
	def __init__(self):
		pass
		
	@staticmethod
	def from_buff(buff):
		length = int.from_bytes(buff.read(4), byteorder = 'big', signed = False)
		if length == 0:
			return []
		data = buff.read(length).decode()
		return data.split(',')
		
	@staticmethod
	def to_bytes(lst):
		if len(lst) == 0:
			return len(lst).to_bytes(4, byteorder="big", signed = False)
		data = ','.join(lst)
		data = data.encode()
		length = len(data).to_bytes(4, byteorder="big", signed = False)
		return length + data
	
	
class SSHMessageNumber(enum.Enum):
	SSH_MSG_DISCONNECT = 1
	SSH_MSG_IGNORE=2
	SSH_MSG_UNIMPLEMENTED=3
	SSH_MSG_DEBUG=4
	SSH_MSG_SERVICE_REQUEST=5
	SSH_MSG_SERVICE_ACCEPT=6

	SSH_MSG_KEXINIT=20
	SSH_MSG_NEWKEYS=21
	
	SSH2_MSG_KEXDH_INIT = 30
	SSH2_MSG_KEXDH_REPLY = 31
	
	SSH_MSG_USERAUTH_REQUEST = 50
	SSH_MSG_USERAUTH_FAILURE = 51
	SSH_MSG_USERAUTH_SUCCESS = 52
	SSH_MSG_USERAUTH_BANNER = 53

class SSHPacket:
	def __init__(self):
		self.packet_length = None
		self.padding_length = None
		self.payload = None
		self.random_padding = None
		self.mac = None
		
	def to_bytes(self, cipher = None):
		if not cipher:
			align = 8
			payload = self.payload.to_bytes()
			padlen = 3 + align - ((len(payload) + 8) % align)
			random_padding = os.urandom(padlen)
			packet_length = len(payload) + len(random_padding)  + 1
			return packet_length.to_bytes(4, byteorder="big", signed = False) + len(random_padding).to_bytes(1, byteorder="big", signed = False) + payload + random_padding
		else:
			align = cipher.get_server_cipher_blocksize()
			payload = self.payload.to_bytes()
			padlen = 3 + align - ((len(payload) + 8) % align)
			random_padding = os.urandom(padlen)
			packet_length = len(payload) + len(random_padding)  + 1
			
			return packet_length.to_bytes(4, byteorder="big", signed = False) + len(random_padding).to_bytes(1, byteorder="big", signed = False) + payload + random_padding
		
	def __repr__(self):
		return str(self)
	
	def __str__(self):
		t = ''
		t += 'length: %s\r\n' % self.packet_length
		t += 'padding length: %s\r\n' % self.padding_length
		t += 'payload: %s\r\n' % str(self.payload)
		t += 'random_padding: %s\r\n' % self.random_padding
		t += 'mac: %s\r\n' % self.mac
		return t
		
class SSH_MSG_KEXINIT:
	def __init__(self):
		self.packet_type = SSHMessageNumber.SSH_MSG_KEXINIT
		self.cookie = None
		self.kex_algorithms = None
		self.server_host_key_algorithms = None
		self.encryption_algorithms_client_to_server = None
		self.encryption_algorithms_server_to_client = None
		self.mac_algorithms_client_to_server = None
		self.mac_algorithms_server_to_client = None
		self.compression_algorithms_client_to_server = None
		self.compression_algorithms_server_to_client = None
		self.languages_client_to_server = None
		self.languages_server_to_client = None
		self.first_kex_packet_follows = None
		self.dummy = 0
		
		self.raw = b'' #raw bytes of the message, will be used later for calculating keys
		
	@staticmethod
	def from_bytes(data):
		return SSH_MSG_KEXINIT.from_buffer(io.BytesIO(data))
		
	@staticmethod
	def from_buffer(buff):
		msg = SSH_MSG_KEXINIT()
		
		msg.raw = buff.read()
		buff = io.BytesIO(msg.raw)
		
		msg.packet_type = SSHMessageNumber(buff.read(1)[0])
		msg.cookie = buff.read(16)
		msg.kex_algorithms = NameList.from_buff(buff)
		msg.server_host_key_algorithms = NameList.from_buff(buff)
		msg.encryption_algorithms_client_to_server = NameList.from_buff(buff)
		msg.encryption_algorithms_server_to_client = NameList.from_buff(buff)
		msg.mac_algorithms_client_to_server = NameList.from_buff(buff)
		msg.mac_algorithms_server_to_client = NameList.from_buff(buff)
		msg.compression_algorithms_client_to_server = NameList.from_buff(buff)
		msg.compression_algorithms_server_to_client = NameList.from_buff(buff)
		msg.languages_client_to_server = NameList.from_buff(buff)
		msg.languages_server_to_client = NameList.from_buff(buff)
		msg.first_kex_packet_follows = bool(int.from_bytes(buff.read(1), byteorder = 'big', signed = False))
		msg.dummy = int.from_bytes(buff.read(4), byteorder = 'big', signed = False) #should be 0
		return msg
		
	def to_bytes(self):
		data = b''
		data += self.packet_type.value.to_bytes(1, byteorder = 'big', signed = False)
		data += self.cookie
		data += NameList.to_bytes(self.kex_algorithms)
		data += NameList.to_bytes(self.server_host_key_algorithms)
		data += NameList.to_bytes(self.encryption_algorithms_client_to_server)
		data += NameList.to_bytes(self.encryption_algorithms_server_to_client)
		data += NameList.to_bytes(self.mac_algorithms_client_to_server)
		data += NameList.to_bytes(self.mac_algorithms_server_to_client)
		data += NameList.to_bytes(self.compression_algorithms_client_to_server)
		data += NameList.to_bytes(self.compression_algorithms_server_to_client)
		data += NameList.to_bytes(self.languages_client_to_server)
		data += NameList.to_bytes(self.languages_server_to_client)
		data += b'\x00' if not self.first_kex_packet_follows else b'\x01'
		data += self.dummy.to_bytes(4, byteorder = 'big', signed = False)

		self.raw = data
		return data
		
	def __str__(self):
		t = ''
		t += 'Packet type: %s\r\n' % self.packet_type.name
		t += 'cookie: %s\r\n' % self.cookie.hex()
		t += 'kex_algorithms: %s\r\n' % ','.join(self.kex_algorithms)
		t += 'server_host_key_algorithms: %s\r\n' % ','.join(self.server_host_key_algorithms)
		t += 'encryption_algorithms_client_to_server: %s\r\n' % ','.join(self.encryption_algorithms_client_to_server)
		t += 'encryption_algorithms_server_to_client: %s\r\n' % ','.join(self.encryption_algorithms_server_to_client)
		t += 'mac_algorithms_client_to_server: %s\r\n' % ','.join(self.mac_algorithms_client_to_server)
		t += 'mac_algorithms_server_to_client: %s\r\n' % ','.join(self.mac_algorithms_server_to_client)
		t += 'compression_algorithms_client_to_server: %s\r\n' % ','.join(self.compression_algorithms_client_to_server)
		t += 'compression_algorithms_server_to_client: %s\r\n' % ','.join(self.compression_algorithms_server_to_client)
		t += 'languages_client_to_server: %s\r\n' % ','.join(self.languages_client_to_server)
		t += 'languages_server_to_client: %s\r\n' % ','.join(self.languages_server_to_client)
		t += 'first_kex_packet_follows: %s\r\n' % str(self.first_kex_packet_follows)
		t += 'dummy: %s\r\n' % str(self.dummy)
		
		return t
		
class SSH2_MSG_KEXDH_INIT:
	def __init__(self):
		self.packet_type = SSHMessageNumber.SSH2_MSG_KEXDH_INIT
		self.e = None #https://tools.ietf.org/html/rfc4253 section 8

	@staticmethod
	def from_bytes(data):
		return SSH2_MSG_KEXDH_INIT.from_buffer(io.BytesIO(data))
		
	@staticmethod
	def from_buffer(buff):
		msg = SSH2_MSG_KEXDH_INIT()
		msg.packet_type = SSHMessageNumber(buff.read(1)[0])
		msg.e = MPInt.from_buff(buff)
		return msg
		
	def to_bytes(self):
		data = b''
		data += self.packet_type.value.to_bytes(1, byteorder = 'big', signed = False)
		data += MPInt.to_bytes(self.e)
		return data
		
	def __str__(self):
		t = ''
		t += 'Packet type: %s\r\n' % self.packet_type.name
		t += 'e: %s\r\n' % self.e
		return t
		
class SSHRSACertData:
	def __init__(self):
		self.identifier = "ssh-rsa" #SSHstring
		self.e = None #mpint
		self.n = None #mpint
	
	@staticmethod
	def from_bytes(data):
		return SSHRSACertData.from_buffer(io.BytesIO(data))
		
	@staticmethod
	def from_buffer(buff):
		buff.read(4)
		msg = SSHRSACertData()
		msg.identifier = SSHstring.from_buff(buff)
		msg.e = MPInt.from_buff(buff)
		msg.n = MPInt.from_buff(buff)
		return msg
		
	def to_bytes(self):
		t = b''
		t += SSHstring.to_bytes(self.identifier)
		t += MPInt.to_bytes(self.e)
		t += MPInt.to_bytes(self.n)
		t = len(t).to_bytes(4, byteorder = 'big', signed = False) + t
		return t
		
	def __str__(self):
		t = 'SSHRSACertData\r\n'
		t += 'identifier: %s\r\n' % self.identifier
		t += 'e: %s\r\n' % self.e
		t += 'n: %s\r\n' % self.n
		return t

class SSHRSASignatureData:
	def __init__(self):
		self.identifier = "ssh-rsa" #SSHstring
		self.rsa_signature_blob = None #bytes of the result of the signing operation
	
	@staticmethod
	def from_bytes(data):
		return SSHRSASignatureData.from_buffer(io.BytesIO(data))
		
	@staticmethod
	def from_buffer(buff):
		buff.read(4)
		msg = SSHRSASignatureData()
		msg.identifier = SSHstring.from_buff(buff)
		msg.rsa_signature_blob = SSHBytes.from_buff(buff)
		return msg
		
	def to_bytes(self):
		t = b''
		t += SSHstring.to_bytes(self.identifier)
		t += SSHBytes.to_bytes(self.rsa_signature_blob)
		t = len(t).to_bytes(4, byteorder = 'big', signed = False) + t
		return t
		
	def __str__(self):
		t = 'SSHRSASignatureData\r\n'
		t += 'identifier: %s\r\n' % self.identifier
		t += 'rsa_signature_blob: %s\r\n' % self.rsa_signature_blob
		return t		

		
class SSHDSSCertData:
	def __init__(self):
		self.identifier = "ssh-dss" #SSHstring
		self.p = None #mpint
		self.q = None #mpint
		self.g = None #mpint
		self.y = None #mpint
	
	@staticmethod
	def from_bytes(data):
		return SSHDSSCertData.from_buffer(io.BytesIO(data))
		
	@staticmethod
	def from_buffer(buff):
		buff.read(4)
		msg = SSHDSSCertData()
		msg.identifier = SSHstring.from_buff(buff)
		self.p = MPInt.from_buff(buff)
		self.q = MPInt.from_buff(buff)
		self.g = MPInt.from_buff(buff)
		self.y = MPInt.from_buff(buff)
		t = len(t).to_bytes(4, byteorder = 'big', signed = False) + t
		return msg
		
	def to_bytes(self):
		t = b''
		t += SSHstring.to_bytes(self.identifier)
		t += MPInt.to_bytes(self.p)
		t += MPInt.to_bytes(self.q)
		t += MPInt.to_bytes(self.g)
		t += MPInt.to_bytes(self.y)
		return t
		
	def __str__(self):
		t = 'SSHDSSCertData\r\n'
		t += 'identifier: %s\r\n' % self.identifier
		t += 'p: %s\r\n' % self.p
		t += 'q: %s\r\n' % self.q
		t += 'g: %s\r\n' % self.g
		t += 'y: %s\r\n' % self.y
		return t
		

		
		
class SSH2_MSG_KEXDH_REPLY:
	def __init__(self):
		self.packet_type = SSHMessageNumber.SSH2_MSG_KEXDH_REPLY
		self.pubkey_string = None
		self.f = None
		self.h_sig = None
		
	@staticmethod
	def from_bytes(data):
		return SSH2_MSG_KEXDH_REPLY.from_buffer(io.BytesIO(data))
		
	@staticmethod
	def from_buffer(buff):
		msg = SSH2_MSG_KEXDH_REPLY()
		msg.packet_type = SSHMessageNumber(buff.read(1)[0])
		msg.pubkey_string = None
		msg.f = None
		msg.h_sig = None
		return msg
		
	def to_bytes(self):
		data = b''
		data += self.packet_type.value.to_bytes(1, byteorder = 'big', signed = False)
		data += self.pubkey_string.to_bytes()
		data += MPInt.to_bytes(self.f)
		data += self.h_sig.to_bytes()
		return data
		
	def __str__(self):
		t = ''
		t += 'Packet type: %s\r\n' % self.packet_type.name
		t += 'mpint: %s\r\n' % self.mpint
		return t
		
		
"""
byte      SSH_MSG_KEXDH_REPLY
      string    server public host key and certificates (K_S)
      mpint     f
      string    signature of H
"""
class SSH_MSG_NEWKEYS:
	def __init__(self):
		self.packet_type = SSHMessageNumber.SSH_MSG_NEWKEYS

	@staticmethod
	def from_bytes(data):
		return SSH_MSG_NEWKEYS.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_buffer(buff):
		msg = SSH_MSG_NEWKEYS()
		msg.packet_type = SSHMessageNumber(buff.read(1)[0])
		return msg

	def to_bytes(self):
		data = b''
		data += self.packet_type.value.to_bytes(1, byteorder = 'big', signed = False)
		return data

class SSH_MSG_SERVICE_REQUEST:
	def __init__(self):
		self.packet_type = SSHMessageNumber.SSH_MSG_SERVICE_REQUEST
		self.service_name = None

	@staticmethod
	def from_bytes(data):
		return SSH_MSG_SERVICE_REQUEST.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_buffer(buff):
		msg = SSH_MSG_SERVICE_REQUEST()
		msg.packet_type = SSHMessageNumber(buff.read(1)[0])
		msg.service_name = SSHstring.from_buff(buff)
		return msg

	def to_bytes(self):
		data = b''
		data += self.packet_type.value.to_bytes(1, byteorder = 'big', signed = False)
		data += SSHstring.to_bytes(self.service_name)
		return data

class SSH_MSG_SERVICE_ACCEPT:
	def __init__(self):
		self.packet_type = SSHMessageNumber.SSH_MSG_SERVICE_ACCEPT
		self.service_name = None

	@staticmethod
	def from_bytes(data):
		return SSH_MSG_SERVICE_REQUEST.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_buffer(buff):
		msg = SSH_MSG_SERVICE_REQUEST()
		msg.packet_type = SSHMessageNumber(buff.read(1)[0])
		msg.service_name = SSHstring.from_buff(buff)
		return msg

	def to_bytes(self):
		data = b''
		data += self.packet_type.value.to_bytes(1, byteorder = 'big', signed = False)
		data += SSHstring.to_bytes(self.service_name)
		return data

class SSH_MSG_USERAUTH_REQUEST_PW_METHOD:
	def __init__(self):
		self.is_pwchange = None
		self.password = None
		self.newpassword = None

	@staticmethod
	def from_bytes(data):
		return SSH_MSG_USERAUTH_REQUEST_PW_METHOD.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_buffer(buff):
		msg = SSH_MSG_USERAUTH_REQUEST_PW_METHOD()
		msg.is_pwchange = bool(int.from_bytes(buff.read(1), byteorder = 'big', signed = False))
		msg.password = SSHstring.from_buff(buff)
		if msg.is_pwchange == True:
			msg.newpassword = SSHstring.from_buff(buff)
		return msg

	def to_bytes(self):
		data = b''
		data += int(self.is_pwchange).to_bytes(1, byteorder = 'big', signed = False)
		data += SSHstring.to_bytes(self.password)
		if self.is_pwchange == True:
			data += SSHstring.to_bytes(self.newpassword)
		return data

	def __str__(self):
		t  = '==== SSH_MSG_USERAUTH_REQUEST_PW_METHOD ====\r\n'
		t += 'is_pwchange : %s\r\n' % self.is_pwchange
		t += 'password : %s\r\n' % self.password
		if self.is_pwchange == True:
			t += 'newpassword : %s\r\n' % self.newpassword
		return t

class SSH_MSG_USERAUTH_REQUEST:
	def __init__(self):
		self.packet_type = SSHMessageNumber.SSH_MSG_USERAUTH_REQUEST
		self.user_name = None
		self.service_name = None
		self.method_name = None
		self.method = None #this can be an object or None

	@staticmethod
	def from_bytes(data):
		return SSH_MSG_USERAUTH_REQUEST.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_buffer(buff):
		msg = SSH_MSG_USERAUTH_REQUEST()
		msg.packet_type = SSHMessageNumber(buff.read(1)[0])
		msg.user_name = SSHstring.from_buff(buff)
		msg.service_name = SSHstring.from_buff(buff)
		msg.method_name = SSHstring.from_buff(buff)
		if msg.method_name == 'password':
			msg.method = SSH_MSG_USERAUTH_REQUEST_PW_METHOD.from_buffer(buff)
		#TODO: the other types https://tools.ietf.org/html/rfc4252#section-6
		return msg

	def to_bytes(self):
		data = b''
		data += self.packet_type.value.to_bytes(1, byteorder = 'big', signed = False)
		data += SSHstring.to_bytes(self.user_name)
		data += SSHstring.to_bytes(self.service_name)
		data += SSHstring.to_bytes(self.service_name)
		data += SSHstring.to_bytes(self.method_name)
		if self.method:
			data += self.method.to_bytes()
		return data

	def __str__(self):
		t  = '==== SSH_MSG_USERAUTH_REQUEST ====\r\n'
		t += 'packet_type : %s\r\n' % self.packet_type
		t += 'user_name : %s\r\n' % self.user_name
		t += 'service_name : %s\r\n' % self.service_name
		t += 'method_name : %s\r\n' % self.method_name
		t += 'method : %s\r\n' % self.method
		return t

class SSH_MSG_USERAUTH_FAILURE:
	def __init__(self):
		self.packet_type = SSHMessageNumber.SSH_MSG_USERAUTH_FAILURE
		self.authentications = []
		self.partial_success = None

	@staticmethod
	def from_bytes(data):
		return SSH_MSG_USERAUTH_FAILURE.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_buffer(buff):
		msg = SSH_MSG_USERAUTH_FAILURE()
		msg.packet_type = SSHMessageNumber(buff.read(1)[0])
		msg.authentications = NameList.from_buff(buff)
		msg.partial_success = bool(int.from_bytes(buff.read(1), byteorder = 'big', signed = False))
		return msg

	def to_bytes(self):
		data = b''
		data += self.packet_type.value.to_bytes(1, byteorder = 'big', signed = False)
		data += NameList.to_bytes(self.authentications)
		data += int(self.partial_success).to_bytes(1, byteorder = 'big', signed = False)
		return data

	def __str__(self):
		t  = '==== SSH_MSG_USERAUTH_FAILURE ====\r\n'
		t += 'authentications : %s\r\n' % ','.join(self.authentications)
		t += 'partial_success : %s\r\n' % self.partial_success
		return t

class SSH_MSG_USERAUTH_SUCCESS:
	def __init__(self):
		self.packet_type = SSHMessageNumber.SSH_MSG_USERAUTH_SUCCESS

	@staticmethod
	def from_bytes(data):
		return SSH_MSG_USERAUTH_SUCCESS.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_buffer(buff):
		msg = SSH_MSG_USERAUTH_SUCCESS()
		msg.packet_type = SSHMessageNumber(buff.read(1)[0])
		return msg

	def to_bytes(self):
		data = b''
		data += self.packet_type.value.to_bytes(1, byteorder = 'big', signed = False)
		return data

	def __str__(self):
		t  = '==== SSH_MSG_USERAUTH_SUCCESS ====\r\n'
		return t

class SSH_MSG_IGNORE:
	def __init__(self):
		self.packet_type = SSHMessageNumber.SSH_MSG_IGNORE
		self.data = None

	@staticmethod
	def from_bytes(data):
		return SSH_MSG_IGNORE.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_buffer(buff):
		msg = SSH_MSG_IGNORE()
		msg.packet_type = SSHMessageNumber(buff.read(1)[0])
		msg.data = SSHBytes.from_buff(buff)
		return msg

	def to_bytes(self):
		data = b''
		data += self.packet_type.value.to_bytes(1, byteorder = 'big', signed = False)
		data += SSHBytes.to_bytes(self.data)
		return data

	def __str__(self):
		t  = '==== SSH_MSG_IGNORE ====\r\n'
		t  = 'data: %s\r\n' % self.data
		return t

class SSH_MSG_DEBUG:
	def __init__(self):
		self.packet_type = SSHMessageNumber.SSH_MSG_DEBUG
		self.always_display = None
		self.message = None
		self.language = None

	@staticmethod
	def from_bytes(data):
		return SSH_MSG_DEBUG.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_buffer(buff):
		msg = SSH_MSG_DEBUG()
		msg.packet_type = SSHMessageNumber(buff.read(1)[0])
		msg.always_display = bool(int.from_bytes(buff.read(1), byteorder = 'big', signed = False))
		msg.message = SSHstring.from_buff(buff)
		msg.language = SSHstring.from_buff(buff)
		return msg

	def to_bytes(self):
		data = b''
		data += self.packet_type.value.to_bytes(1, byteorder = 'big', signed = False)
		data += int(self.always_display).to_bytes(1, byteorder = 'big', signed = False)
		data += SSHstring.to_bytes(self.message)
		data += SSHstring.to_bytes(self.language)
		return data

	def __str__(self):
		t  = '==== SSH_MSG_DEBUG ====\r\n'
		t  = 'always_display: %s\r\n' % self.always_display
		t  = 'message: %s\r\n' % self.message
		t  = 'language: %s\r\n' % self.language
		return t

class SSH_MSG_UNIMPLEMENTED:
	def __init__(self):
		self.packet_type = SSHMessageNumber.SSH_MSG_UNIMPLEMENTED
		self.sequence_number = None

	@staticmethod
	def from_bytes(data):
		return SSH_MSG_UNIMPLEMENTED.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_buffer(buff):
		msg = SSH_MSG_UNIMPLEMENTED()
		msg.sequence_number = int.from_bytes(buff.read(4), byteorder = 'big', signed = False)
		return msg

	def to_bytes(self):
		data = b''
		data += self.packet_type.value.to_bytes(1, byteorder = 'big', signed = False)
		data += self.sequence_number.to_bytes(4, byteorder = 'big', signed = False)
		return data

	def __str__(self):
		t  = '==== SSH_MSG_UNIMPLEMENTED ====\r\n'
		t  = 'sequence_number: %s\r\n' % self.sequence_number
		return t

class SSH_MSG_DISCONNECT:
	def __init__(self):
		self.packet_type = SSHMessageNumber.SSH_MSG_DISCONNECT
		self.reason_code = None
		self.description = None
		self.language = None

	@staticmethod
	def from_bytes(data):
		return SSH_MSG_DISCONNECT.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_buffer(buff):
		msg = SSH_MSG_DISCONNECT()
		msg.reason_code = int.from_bytes(buff.read(4), byteorder = 'big', signed = False)
		msg.always_display = bool(int.from_bytes(buff.read(1), byteorder = 'big', signed = False))
		msg.description = SSHstring.from_buff(buff)
		msg.language = SSHstring.from_buff(buff)
		return msg

	def to_bytes(self):
		data = b''
		data += self.packet_type.value.to_bytes(1, byteorder = 'big', signed = False)
		data += self.reason_code.to_bytes(1, byteorder = 'big', signed = False)
		data += SSHstring.to_bytes(self.description)
		data += SSHstring.to_bytes(self.language)
		return data

	def __str__(self):
		t  = '==== SSH_MSG_DISCONNECT ====\r\n'
		t  = 'reason_code: %s\r\n' % self.reason_code
		t  = 'description: %s\r\n' % self.description
		t  = 'language: %s\r\n' % self.language
		return t

type2msg = {
	SSHMessageNumber.SSH_MSG_DISCONNECT : SSH_MSG_DISCONNECT, 
	SSHMessageNumber.SSH_MSG_IGNORE : SSH_MSG_IGNORE, 
	SSHMessageNumber.SSH_MSG_UNIMPLEMENTED : SSH_MSG_UNIMPLEMENTED, 
	SSHMessageNumber.SSH_MSG_DEBUG : SSH_MSG_DEBUG, 
	SSHMessageNumber.SSH_MSG_SERVICE_REQUEST : SSH_MSG_SERVICE_REQUEST, 
	SSHMessageNumber.SSH_MSG_SERVICE_ACCEPT : SSH_MSG_SERVICE_ACCEPT, 
	SSHMessageNumber.SSH_MSG_KEXINIT : SSH_MSG_KEXINIT, 
	SSHMessageNumber.SSH_MSG_NEWKEYS : SSH_MSG_NEWKEYS,
	SSHMessageNumber.SSH2_MSG_KEXDH_INIT : SSH2_MSG_KEXDH_INIT,
	SSHMessageNumber.SSH2_MSG_KEXDH_REPLY : SSH2_MSG_KEXDH_REPLY,
	SSHMessageNumber.SSH_MSG_USERAUTH_REQUEST : SSH_MSG_USERAUTH_REQUEST,
	SSHMessageNumber.SSH_MSG_USERAUTH_FAILURE : SSH_MSG_USERAUTH_FAILURE,
	SSHMessageNumber.SSH_MSG_USERAUTH_SUCCESS : SSH_MSG_USERAUTH_SUCCESS,
}


dh_groups = {
	# 2048-bit
	14: {
		"p": 0xFFFFFFFFFFFFFFFFC90FDAA22168C234C4C6628B80DC1CD129024E088A67CC74020BBEA63B139B22514A08798E3404DDEF9519B3CD3A431B302B0A6DF25F14374FE1356D6D51C245E485B576625E7EC6F44C42E9A637ED6B0BFF5CB6F406B7EDEE386BFB5A899FA5AE9F24117C4B1FE649286651ECE45B3DC2007CB8A163BF0598DA48361C55D39A69163FA8FD24CF5F83655D23DCA3AD961C62F356208552BB9ED529077096966D670C354E4ABC9804F1746C08CA18217C32905E462E36CE3BE39E772C180E86039B2783A2EC07A28FB5C55DF06F4C52C9DE2BCBF6955817183995497CEA956AE515D2261898FA051015728E5A8AACAA68FFFFFFFFFFFFFFFF,
		"g": 2, 
		"keysize" : 2048
	},
}

class DH:
	def __init__(self, group_no, privkey = None):
		self.p = dh_groups[group_no]['p']
		self.g = dh_groups[group_no]['g']
		self.keysize = int(dh_groups[group_no]['keysize'] / 8)
		
		maxin = (self.p - 1) // 2
		bs = len(maxin.to_bytes((maxin.bit_length() + 7) // 8, 'big'))
		self.y = int.from_bytes(os.urandom(bs), byteorder = 'big', signed = False) if not privkey else privkey		
		self.f = pow(self.g, self.y, self.p) # dh_server_pub
		self.shared_key = None
		
	def get_shared_key(self, e):
		if not (2 <= e and e <= self.p - 2):
			raise Exception('Remote key invalid!')
			
		#print('DH e: %s' % e)
		#print('DH y: %s' % self.y)
		#print('DH p: %s' % self.p)
		
		self.shared_key = pow(e, self.y, self.p)
		return self.shared_key
		#return hashlib.sha256(self.shared_key.to_bytes(256, byteorder = 'big', signed = False)).hexdigest()
	
class SSHCipher:
	def __init__(self, privkey_file):	
		self.server_host_keys = {
			SSHPrivKey.name : SSHPrivKey.load_privkey_from_file(privkey_file)
		}
		
		"""
		self.kex_algorithms = [
			
		]
		"""
	
		# banner and KEX messages can be hardcoded. pick diffie-hellman-group14-sha1, aes128-ctr, hmac-sha1 or sha256, no compression
		self.supported_kex_algorithms = ['diffie-hellman-group14-sha1']
		self.supported_server_host_key_algorithms = [x for x in self.server_host_keys]
		self.supported_encryption_algorithms = ['aes128-ctr']
		self.supported_mac_algorithms = ['hmac-sha1']
		self.supported_compression_algorithms = ['none']
		self.supported_languages = []
		
		self.selected_kex_algorithm = 'diffie-hellman-group14-sha1'
		self.selected_server_host_key_algorithm = 'ssh-rsa'
		self.selected_encryption_algorithm = 'aes128-ctr'
		self.selected_mac_algorithm = 'hmac-sha1'
		self.selected_compression_algorithm = None
		self.selected_language = None
		
		#self.server_rsa_key = 
		#self.server_dsa_key = None
		#self.server_key = None #this should be set via server starup config parameters
		#self.pubkey, self.privkey = rsa.newkeys(2048)
		#self.server_kex_assymmetric =   #this should be set automatically but currently we only use one
		
		#self.kex = None #this should be set automatically but currently we only use one
		self.kex = DH(14)
		self.server_cipher = None
		self.client_cipher = None
		self.server_hmac = None
		self.client_hmac = None
		self.selected_mac_algorithm_obj = hashlib.sha1

	def get_server_cipher_blocksize(self):
		return 16

	def get_server_hmac(self):
		return hmac.new(self.s2c_integrity_key, digestmod=self.selected_mac_algorithm_obj)

	def get_client_hmac(self):
		return hmac.new(self.c2s_integrity_key, digestmod=self.selected_mac_algorithm_obj)
		
	def generate_server_key_rply(self):
		msg = SSH_MSG_KEXINIT()
		msg.cookie = os.urandom(16)
		msg.kex_algorithms = self.supported_kex_algorithms
		msg.server_host_key_algorithms = self.supported_server_host_key_algorithms
		msg.encryption_algorithms_client_to_server = self.supported_encryption_algorithms
		msg.encryption_algorithms_server_to_client = self.supported_encryption_algorithms
		msg.mac_algorithms_client_to_server = self.supported_mac_algorithms
		msg.mac_algorithms_server_to_client = self.supported_mac_algorithms
		msg.compression_algorithms_client_to_server = self.supported_compression_algorithms
		msg.compression_algorithms_server_to_client = self.supported_compression_algorithms
		msg.languages_client_to_server = self.supported_languages
		msg.languages_server_to_client = self.supported_languages
		msg.first_kex_packet_follows = False
		return msg
		
	
		

	def calculate_kex(self, client_id, server_id, client_kexinit_payload, server_kexinit_payload, client_msg):
		#currently only DH + RSA is supported!
		
		###### SHARED SECRET CALC
		if self.selected_kex_algorithm == 'diffie-hellman-group14-sha1':
			"""
			#parameters = dh.generate_parameters(generator=2, key_size=2048, backend=default_backend())
			# Generate a private key for use in the exchange.
			sc = pow(client_msg.e, self.kex.y, self.kex.p)
			K = MPInt.int2bytes(sc, False)
			self.kex_hash = hashlib.sha1
			
			#calculating client's secret key
			#sc = self.kex.get_shared_key(client_msg.e)
			#K =  MPInt.int2bytes(sc, False) #shared secret
			"""
			self.kex_hash = hashlib.sha1
			self.p = 0xFFFFFFFFFFFFFFFFC90FDAA22168C234C4C6628B80DC1CD129024E088A67CC74020BBEA63B139B22514A08798E3404DDEF9519B3CD3A431B302B0A6DF25F14374FE1356D6D51C245E485B576625E7EC6F44C42E9A637ED6B0BFF5CB6F406B7EDEE386BFB5A899FA5AE9F24117C4B1FE649286651ECE45B3DC2007CB8A163BF0598DA48361C55D39A69163FA8FD24CF5F83655D23DCA3AD961C62F356208552BB9ED529077096966D670C354E4ABC9804F1746C08CA18217C32905E462E36CE3BE39E772C180E86039B2783A2EC07A28FB5C55DF06F4C52C9DE2BCBF6955817183995497CEA956AE515D2261898FA051015728E5A8AACAA68FFFFFFFFFFFFFFFF
			self.g = 2
			self.x = 31337
			self.f = pow(self.g, self.x, self.p)
			sc = pow(client_msg.e, self.x, self.p)
			#K = MPInt.int2bytes(sc)
			K = MPInt.int2bytes_padded(sc)
			
			
			#print('Shared Secret -int-: %s' % sc)
			#print('Shared Secret -bytes-: %s' % K.hex())
			
		
		else:
			raise Exception('Unsupported KEX algo!')
			
		if self.selected_server_host_key_algorithm == 'ssh-rsa':
			pkey = self.server_host_keys['ssh-rsa']
			#print('Server host key "e" %s' % pkey.privkey.e)
			#print('Server host key "N" %s' % pkey.privkey.n)
			server_host_key = SSHRSACertData()
			server_host_key.e = pkey.privkey.e
			server_host_key.n = pkey.privkey.n
		
		else:
			raise Exception('Server host key algo!')
		
		#print('Client id: %s' % client_id)
		#print('server_id: %s' % server_id)
		#print('client_kexinit_payload: \r\n%s' % hexdump(client_kexinit_payload))
		#print('server_kexinit_payload: \r\n%s' % hexdump(server_kexinit_payload))
		#print('server_host_key:\r\n %s' % hexdump((server_host_key.to_bytes())))
		#print('client_msg.e: %s' % MPInt.int2bytes(client_msg.e).hex())
		#print('dh_server_pub: %s' % MPInt.int2bytes(self.f).hex())
		#print('K: %s' % K)
		
		hash_buffer = b''
		hash_buffer += SSHBytes.to_bytes(client_id)
		#print('client_id end : %s' % hex(len(hash_buffer)))
		hash_buffer += SSHBytes.to_bytes(server_id)
		#print('server_id end : %s' % hex(len(hash_buffer)))
		hash_buffer += SSHBytes.to_bytes(client_kexinit_payload)
		#print('client_kexinit_payload end : %s' % hex(len(hash_buffer)))
		hash_buffer += SSHBytes.to_bytes(server_kexinit_payload)
		#print('server_kexinit_payload end : %s' % hex(len(hash_buffer)))
		hash_buffer += server_host_key.to_bytes()
		#print('server_host_key end : %s' % hex(len(hash_buffer)))
		hash_buffer += SSHBytes.to_bytes(MPInt.int2bytes_padded(client_msg.e))
		#print('client_msg.e end : %s' % hex(len(hash_buffer)))
		hash_buffer += SSHBytes.to_bytes(MPInt.int2bytes_padded(self.f))
		#print('self.f end : %s' % hex(len(hash_buffer)))
		hash_buffer += SSHBytes.to_bytes(K)
		
		#print(hexdump(hash_buffer))
		
		#print('Hash buffer: %s' % hash_buffer.hex())
		
		H = self.kex_hash(hash_buffer).digest()#H = hash(V_C || V_S || I_C || I_S || K_S || e || f || K)
		session_id =  H
		#print('H: %s' % H.hex())
		
		K_enc = SSHBytes.to_bytes(K)
		self.c2s_init_IV = self.kex_hash(K_enc + H + b'A' + session_id).digest() # Initial IV client to server: HASH(K || H || "A" || session_id)
		self.s2c_init_IV = self.kex_hash(K_enc + H + b'B' + session_id).digest() # o   Initial IV server to client: HASH(K || H || "B" || session_id)
		self.c2s_cipher_key = self.kex_hash(K_enc + H + b'C' + session_id).digest() #o  Encryption key client to server: HASH(K || H || "C" || session_id)
		self.s2c_cipher_key = self.kex_hash(K_enc + H + b'D' + session_id).digest() #o  Encryption key server to client: HASH(K || H || "D" || session_id)
		self.c2s_integrity_key = self.kex_hash(K_enc + H + b'E' + session_id).digest() #o  Integrity key client to server: HASH(K || H || "E" || session_id)
		self.s2c_integrity_key = self.kex_hash(K_enc + H + b'F' + session_id).digest() #o  Integrity key server to client: HASH(K || H || "F" || session_id)
		
		#print('A: %s' % self.c2s_init_IV.hex())
		#print('B: %s' % self.s2c_init_IV.hex())
		#print('C: %s' % self.c2s_cipher_key.hex())
		#print('D: %s' % self.s2c_cipher_key.hex())
		#print('E: %s' % self.c2s_integrity_key.hex())
		#print('F: %s' % self.s2c_integrity_key.hex())

		self.server_cipher = AESModeOfOperationCTR(self.s2c_cipher_key[:16], Counter(int.from_bytes(self.s2c_init_IV[:16], byteorder = 'big', signed = False) ))
		self.client_cipher = AESModeOfOperationCTR(self.c2s_cipher_key[:16], Counter(int.from_bytes(self.c2s_init_IV[:16], byteorder = 'big', signed = False) ))
		self.server_hmac = hmac.new(self.s2c_integrity_key, digestmod=hashlib.sha1)
		self.client_hmac = hmac.new(self.c2s_integrity_key, digestmod=hashlib.sha1)
		
		h_sig = SSHRSASignatureData()
		h_sig.rsa_signature_blob = pkey.sign(H)
		
		payload = SSH2_MSG_KEXDH_REPLY()
		payload.pubkey_string = server_host_key
		payload.f = self.f
		payload.h_sig = h_sig
		
		return payload
		
		
		
class SSHParser:
	def __init__(self):
		pass
	
	@staticmethod
	async def from_streamreader(reader, cipher = None, sequence_number = None):
			packet = SSHPacket()
			if not cipher:
				#packet is not encrypted at this point
				packet_length = await readexactly_or_exc(reader, 4)
				packet.packet_length = int.from_bytes(packet_length, byteorder = 'big', signed = False)
				if packet.packet_length > MAX_PACKET_LENGTH:
					raise Exception('SSH packet too large!')
					
				data = await readexactly_or_exc(reader, packet.packet_length)
				packet.padding_length = data[0]
				message_type = SSHMessageNumber(data[1])
				#print(message_type)
				packet.payload = type2msg[message_type].from_bytes(data[1:(packet.packet_length - packet.padding_length)])
				packet.random_padding = data[-packet.padding_length:]
				packet.mac = None
				return packet
			else:
				blocksize = cipher.get_server_cipher_blocksize() #get this info from teh cipher!!!
				packet_length_enc = await readexactly_or_exc(reader, blocksize)
				packet_length_dec = cipher.client_cipher.decrypt(packet_length_enc)
				packet.packet_length = int.from_bytes(packet_length_dec[:4], byteorder = 'big', signed = False)
				if packet.packet_length > MAX_PACKET_LENGTH:
					raise Exception('SSH packet too large!')
				data_enc = await readexactly_or_exc(reader, packet.packet_length - blocksize + 4)
				data = packet_length_dec[4:] + cipher.client_cipher.decrypt(data_enc)
				data = io.BytesIO(data)
				packet.padding_length = data.read(1)[0]
				message_type = SSHMessageNumber( data.read(1)[0])
				data.seek(-1, 1)
				packet.payload = type2msg[message_type].from_bytes(data.read(packet.packet_length - packet.padding_length - 1))
				packet.random_padding = data.read()
				packet.mac = await readexactly_or_exc(reader, cipher.client_hmac.digest_size)

				#mac = MAC(key, sequence_number || unencrypted_packet)
				data.seek(0, 0)
				unencrypted_packet = packet_length_dec[:4] + data.read()
				a = cipher.get_client_hmac()
				a.update(sequence_number +  unencrypted_packet)
				if not hmac.compare_digest(a.digest(), packet.mac):
					raise Exception('Client packet HMAC error!')

				return packet
			
	@staticmethod
	def from_bytes(data):
		return SSHParser.from_buffer(io.BytesIO(data))	
	
	@staticmethod
	def from_buffer(buff):
		packet = SSHPacket()
		packet.packet_length = int.from_bytes(buff.read(4), byteorder = 'big', signed = False)
		data = buff.read(packet.packet_length)
		packet.padding_length = data[0]
		message_type = SSHMessageNumber(data[1])
		packet.payload = type2msg[message_type].from_bytes(data[1:(packet.packet_length - packet.padding_length)])
		packet.random_padding = data[-packet.padding_length:]
		packet.mac = None
		return packet
		


#https://msdn.microsoft.com/en-us/library/dd304523.aspx
#https://msdn.microsoft.com/en-us/library/ee209073(v=sql.105).aspx

#This is the implamentation of the TDS protocol
#The TDS protocol does not prescribe a specific underlying transport protocol to use on the Internet or on other networks. TDS only presumes a reliable transport that guarantees in-sequence delivery of data.

#TODO
#the actual implementation is up for the imagination of the reader

import enum
import io

from responder3.core.asyncio_helpers import *

class SessionStatus(enum.Enum):
	START = enum.auto()
	SSL_AUTH = enum.auto()
	LOGIN7_AUTH = enum.auto()
	SSPI_AUTH = enum.auto()
	AUHTENTICATED = enum.auto()
	PREAUTH_SENT = enum.auto()

class Encryption(enum.Enum):
	OFF = 0x00 #Encryption is available but off.
	ON = 0x01 #Encryption is available and on.
	NOT_SUP = 0x02 #Encryption is not available.
	REQ = 0x03 #Encryption is required.

class PacketType(enum.Enum):
	SQL_BATCH = 1
	PRE_TDS7_LOGIN = 2
	RPC = 3
	TABULAR_RESULT = 4
	UNUSED_5 = 5
	ATTENTION_SIGNAL =6
	BULK_LOAD_DATA=7
	FEDERATED_AUTH_TOKEN = 8
	UNUSED_9 = 9
	UNUSED_10 = 10
	UNUSED_11 = 11
	UNUSED_12 = 12
	UNUSED_13 = 13
	TRANSACTION_MANAGER_REQ = 14
	UNUSED_15 = 15
	LOGIN7 = 16
	SSPI = 17
	PRELOGIN = 18

class PacketStatus(enum.IntFlag):
	NORMAL = 0x00
	EOM = 0x01
	IGNORE = 0x02
	RESET_CONN = 0x08
	RESET_CONN_SKIP_TRAN = 0x10

class TDSPacket:
	def __init__(self):
		self.type = None
		self.status = None
		self.length = None
		self.spid = None
		self.packet_id = None
		self.window = None
		self.data = None

	def hdr_from_bytes(bbuff):
		return TDSPacket.hdr_from_buffer(io.BytesIO(bbuff))

	def hdr_from_buffer(buffer):
		p = TDSPacket()
		p.type = PacketType(int.from_bytes(buffer.read(1), byteorder = 'big', signed = False))
		p.status = PacketStatus(int.from_bytes(buffer.read(1), byteorder = 'big', signed = False))
		p.length = int.from_bytes(buffer.read(2), byteorder = 'big', signed = False)
		p.spid = int.from_bytes(buffer.read(2), byteorder = 'big', signed = False)
		p.packet_id = int.from_bytes(buffer.read(1), byteorder = 'big', signed = False)
		p.window = int.from_bytes(buffer.read(1), byteorder = 'big', signed = False)
		return p

	def to_bytes(self):
		data = self.data.to_bytes()
		t  = self.type.value.to_bytes(1, byteorder = 'big', signed = False)
		t += self.status.value.to_bytes(1, byteorder = 'big', signed = False)
		t += (len(data)+8).to_bytes(2, byteorder = 'big', signed = False)
		t += self.spid.to_bytes(2, byteorder = 'big', signed = False)
		t += self.packet_id.to_bytes(1, byteorder = 'big', signed = False)
		t += self.window.to_bytes(1, byteorder = 'big', signed = False)
		return t + data

	def data_from_bytes(self, bbuff):
		self.data = TDSData.from_bytes(type2class[self.type],bbuff)

	def data_from_buffer(self, buffer):
		self.data = TDSData.from_buffer(type2class[self.type],bbuff)

	@staticmethod
	async def from_streamreader(reader, timeout=None):
		t_hdr = await readexactly_or_exc(reader, 8, timeout=timeout)
		hdr = TDSPacket.hdr_from_bytes(t_hdr)
		t_data = await readexactly_or_exc(reader, hdr.length - 8, timeout=timeout)
		hdr.data_from_bytes(t_data)
		return hdr

	def __str__(self):
		t = '== TDSPacket ==\r\n'
		t += 'type : %s\r\n' % self.type
		t += 'status : %s\r\n' % self.status
		t += 'length : %s\r\n' % self.length
		t += 'spid : %s\r\n' % self.spid
		t += 'packet_id : %s\r\n' % self.packet_id
		t += 'window : %s\r\n' % self.window
		t += str(self.data)
		return t

class TDSData:
	def __init__(self):
		pass

	def get_tds_fields(buffer):
		pos = buffer.tell()
		buffer.seek(0,0)
		start = buffer.tell()
		buffer.seek(0,2)
		end = buffer.tell()
		buffer_len = end - start
		buffer.seek(pos, 0)
		fields = []
		cur = TDSField()
		while cur.type != 0xff and buffer.tell() < buffer_len:
			cur = TDSField.from_buffer(buffer)
			fields.append(cur)

		return fields

	@staticmethod
	def from_bytes(ptype, bbuff):
		return ptype.from_buffer(ptype, io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(ptype, buffer):
		p = ptype()
		types, inv, field_types = ptype.get_types()
		for field in TDSData.get_tds_fields(buffer):
			if field.type in inv:
				filed_name = inv[field.type]
				field_type = field_types[filed_name]
				if field_type[0] == 'int':
					setattr(p, filed_name, int.from_bytes(field.data, byteorder = field_type[2], signed = field_type[3]))
				elif field_type[0] == 'enum':
					setattr(p, filed_name, field_type[2](int.from_bytes(field.data, byteorder = 'big', signed = False)))
				elif field_type[0] == 'bool':
					setattr(p, filed_name, bool(int.from_bytes(field.data, byteorder = 'big', signed = False)))
				elif field_type[0] == 'byte':
					setattr(p, filed_name, field.data)
				elif field_type[0] == 'str':
					setattr(p, filed_name, field.data.decode())
				else:
					raise Exception('Unknown field type!')

			else:
				if field.type == 0xff:
					continue
				print('Got type %s, but its not defined' % field.type)

		return p

	def to_bytes(self):
		types, inv, field_types = type(self).get_types()
		tfs = []
		for key in types:
			tf = TDSField()
			tf.type = types[key]
			field_type = field_types[key]
			field_data = getattr(self, key)
			if not field_data:
				continue
			if field_type[0] == 'int':
				tf.data = field_data.to_bytes(field_type[1], byteorder = field_type[2], signed = field_type[3])
			elif field_type[0] == 'enum':
				tf.data = field_data.value.to_bytes(field_type[1], byteorder = 'big', signed = False)
			elif field_type[0] == 'bool':
				tf.data = int(field_data).to_bytes(field_type[1], byteorder = 'big', signed = False)
			elif field_type[0] == 'byte':
				tf.data = field_data
			elif field_type[0] == 'str':
				tf.data = field_data.encode() + b'\x00'
			else:
				raise Exception('Unknown field type!')
			tf.length = len(tf.data)
			tf.offset = 0
			tfs.append(tf)
		data_offset = (len(tfs) * 5) +1
		data = b''
		t = b''
		i = 0
		for tf in tfs:
			tf.offset = data_offset + i
			t += tf.to_bytes()
			i += tf.length
			data += tf.data

		return t + b'\xff'+ data

	def __str__(self):
		types, inv, field_type = type(self).get_types()
		t = '=== %s === \r\n' % type(self)
		for key in types:
			t += '%s: %s\r\n' % (key, getattr(self, key))
		return t

# https://msdn.microsoft.com/en-us/library/dd357559.aspx
class PRELOGIN(TDSData):
	def __init__(self):
		TDSData.__init__(self)
		self.version = None
		self.subbuid = None
		self.encryption = None
		self.instvalidity = None
		self.thread_id = None
		self.mars = None
		self.traceid = None
		self.activity_id = None
		self.activity_seq = None
		self.fedauthrequired = None
		self.nonce = None
		self.terminator = None

	@staticmethod
	def get_types():
		t = {
			'version' : 0,
			#'subbuid' : 0,
			'encryption' : 1,
			'instvalidity' : 2,
			'thread_id' : 3,
			'mars' : 4,
			'traceid' : 5,
			#'activity_id' : 0,
			#'activity_seq' : 0,
			'fedauthrequired' : 6,
			'nonce' : 7,
		}
		inv = {v: k for k, v in t.items()}
		filed_types = {
			'version' : ('byte', 6),#('int', 8, 'big', False),
			#'subbuid' : 0,
			'encryption' : ('enum', 1, Encryption),
			'instvalidity' : ('str', -1),
			'thread_id' : ('int', 8, 'big', False),
			'mars' : ('bool', 1),
			'traceid' : ('byte', 40),
			#'activity_id' : 0,
			#'activity_seq' : 0,
			'fedauthrequired' : ('bool', 1),
			'nonce' : ('byte', 32),
		}
		return t, inv, filed_types

class TDSField:
	def __init__(self):
		self.type = None
		self.offset = None
		self.length = None
		self.data = None

	def to_bytes(self):
		t  = self.type.to_bytes(1, byteorder = 'big', signed = False)
		if self.type == 0xff:
			return t
		t += self.offset.to_bytes(2, byteorder = 'big', signed = False)
		t += self.length.to_bytes(2, byteorder = 'big', signed = False)
		return t

	def read_data(self, buffer):
		pos = buffer.tell()
		buffer.seek(self.offset, 0)
		data = buffer.read(self.length)
		buffer.seek(pos, 0)
		return data

	@staticmethod
	def from_bytes(bbuff):
		return TDSField.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buffer):
		tf = TDSField()
		pos_start = buffer.tell()
		tf.type = int.from_bytes(buffer.read(1), byteorder = 'big', signed = False)
		if tf.type == 0xff: #terminator type
			return tf
		tf.offset = int.from_bytes(buffer.read(2), byteorder = 'big', signed = False)
		tf.length = int.from_bytes(buffer.read(2), byteorder = 'big', signed = False)
		tf.data = tf.read_data(buffer)
		return tf

class TDSEndField(TDSField):
	def __init__(self):
		self.type = 0xff


class LOGIN7:
	def __init__(self):
		self.total_length = None
		self.tds_version = None
		self.packet_size = None
		self.clientprogver = None
		self.client_pid = None
		self.connection_id = None
		self.byteorder = None
		self.flags1 = None
		self.flags2 = None
		self.sql_type_flags = None
		self.flags3 = None
		self.timezone = None
		self.collation = None

		self.cliname_off = None
		self.cliname_len = None
		self.username_off = None
		self.username_len = None
		self.password_off = None
		self.password_len = None
		self.appname_off = None
		self.appname_len = None
		self.servername_off = None
		self.servername_len = None
		self.unused_1_off = None
		self.unused_1_len = None
		self.libname_off = None
		self.libname_len = None
		self.locale_off = None
		self.locale_len = None

		self.dbname_off = None
		self.dbname_off = None

		self.cliname = None
		self.username = None
		self.password = None
		self.appname = None
		self.servername = None
		self.unused = None
		self.libname = None
		self.locale = None
		self.dbname = None

	@staticmethod
	def decode_password(enc_password):
		pw = b''
		for c in enc_password:
			temp = c ^ 0xA5
			temp = ((temp << 4)&0xff) | ((temp >> 4) &0xff)
			pw += temp.to_bytes(1, byteorder ='big', signed = False)
		return pw.decode('utf-16le')

	@staticmethod
	def encode_password(password):
		enc_pw = ''
		for c in password.encode('utf-16le'):
			temp = ord(c)
			temp = ((temp << 4)&0xff) | ((temp >> 4) &0xff)
			enc_pw += (temp ^ 0xA5).to_bytes(1, byteorder = 'big', signed = False) 
		return enc_pw

	@staticmethod
	def from_bytes(x, bbuff):
		return LOGIN7.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(x, buffer):
		login = LOGIN7()
		login.total_length = int.from_bytes(buffer.read(4), byteorder = 'little', signed = 'False')
		login.tds_version = int.from_bytes(buffer.read(4), byteorder = 'little', signed = 'False')
		login.packet_size = int.from_bytes(buffer.read(4), byteorder = 'little', signed = 'False')
		login.clientprogver = int.from_bytes(buffer.read(4), byteorder = 'little', signed = 'False')
		login.client_pid = int.from_bytes(buffer.read(4), byteorder = 'little', signed = 'False')
		login.connection_id = int.from_bytes(buffer.read(4), byteorder = 'little', signed = 'False')
		login.flags1 = int.from_bytes(buffer.read(1), byteorder = 'little', signed = 'False')
		login.flags2 = int.from_bytes(buffer.read(1), byteorder = 'little', signed = 'False')
		login.sql_type_flags = int.from_bytes(buffer.read(1), byteorder = 'little', signed = 'False')
		login.flags3 = int.from_bytes(buffer.read(1), byteorder = 'little', signed = 'False')
		login.timezone = int.from_bytes(buffer.read(4), byteorder = 'little', signed = 'False')
		login.collation = int.from_bytes(buffer.read(4), byteorder = 'little', signed = 'False')

		login.cliname_off = int.from_bytes(buffer.read(2), byteorder = 'little', signed = 'False')
		login.cliname_len = int.from_bytes(buffer.read(2), byteorder = 'little', signed = 'False')
		login.username_off = int.from_bytes(buffer.read(2), byteorder = 'little', signed = 'False')
		login.username_len = int.from_bytes(buffer.read(2), byteorder = 'little', signed = 'False')
		login.password_off = int.from_bytes(buffer.read(2), byteorder = 'little', signed = 'False')
		login.password_len = int.from_bytes(buffer.read(2), byteorder = 'little', signed = 'False')
		login.appname_off = int.from_bytes(buffer.read(2), byteorder = 'little', signed = 'False')
		login.appname_len = int.from_bytes(buffer.read(2), byteorder = 'little', signed = 'False')
		login.servername_off = int.from_bytes(buffer.read(2), byteorder = 'little', signed = 'False')
		login.servername_len = int.from_bytes(buffer.read(2), byteorder = 'little', signed = 'False')
		login.unused_1_off = int.from_bytes(buffer.read(2), byteorder = 'little', signed = 'False')
		login.unused_1_len = int.from_bytes(buffer.read(2), byteorder = 'little', signed = 'False')
		login.libname_off = int.from_bytes(buffer.read(2), byteorder = 'little', signed = 'False')
		login.libname_len = int.from_bytes(buffer.read(2), byteorder = 'little', signed = 'False')
		login.locale_off = int.from_bytes(buffer.read(2), byteorder = 'little', signed = 'False')
		login.locale_len = int.from_bytes(buffer.read(2), byteorder = 'little', signed = 'False')

		login.dbname_off = int.from_bytes(buffer.read(2), byteorder = 'little', signed = 'False')
		login.dbname_len = int.from_bytes(buffer.read(2), byteorder = 'little', signed = 'False')
		
		if login.cliname_off != 0:
			buffer.seek(login.cliname_off,0)
			login.cliname = buffer.read(login.cliname_len * 2).decode('utf-16le')
		
		if login.username_off != 0:
			buffer.seek(login.username_off,0)
			login.username = buffer.read(login.username_len * 2).decode('utf-16le')
		
		if login.password_off != 0:
			buffer.seek(login.password_off,0)
			login.password = LOGIN7.decode_password(buffer.read(login.password_len * 2))
		
		if login.appname_off != 0:
			buffer.seek(login.appname_off,0)
			login.appname = buffer.read(login.appname_len * 2).decode('utf-16le')
		
		if login.servername_off != 0:
			buffer.seek(login.servername_off,0)
			login.servername = buffer.read(login.servername_len * 2).decode('utf-16le')

		if login.unused_1_off != 0:
			buffer.seek(login.unused_1_off,0)
			login.unused = buffer.read(login.unused_1_len)
		
		if login.libname_off != 0:
			buffer.seek(login.libname_off,0)
			login.libname = buffer.read(login.libname_len * 2).decode('utf-16le')
		if login.locale_off != 0:
			buffer.seek(login.locale_off,0)
			login.locale = buffer.read(login.locale_len * 2).decode('utf-16le')
		if login.dbname_off != 0:
			buffer.seek(login.dbname_off,0)
			login.dbname = buffer.read(login.dbname_len * 2).decode('utf-16le')

		return login

	def __str__(self):
		return '%s : %s' % (self.username, self.password)

type2class = {
	#PacketType.SQL_BATCH : None,
	#PacketType.PRE_TDS7_LOGIN : None,
	#PacketType.RPC : None,
	#PacketType.TABULAR_RESULT : None,
	#PacketType.UNUSED_5 : None,
	#PacketType.ATTENTION_SIGNAL : None,
	#PacketType.BULK_LOAD_DATA : None,
	#PacketType.FEDERATED_AUTH_TOKEN : None,
	#PacketType.UNUSED_9 : None,
	#PacketType.UNUSED_10 : None,
	#PacketType.UNUSED_11 : None,
	#PacketType.UNUSED_12 : None,
	#PacketType.UNUSED_13 : None,
	#PacketType.TRANSACTION_MANAGER_REQ : None,
	#PacketType.UNUSED_15 : None,
	PacketType.LOGIN7 : LOGIN7,
	#PacketType.SSPI : None,
	PacketType.PRELOGIN : PRELOGIN,
}
type2class_inv = {v: k for k, v in type2class.items()}


import io
import enum
import binascii
import sys
from responder3.protocols.SMB.ntstatus import *
from responder3.protocols.SMB.utils import *


# https://msdn.microsoft.com/en-us/library/ee441774.aspx
class SMBHeader:
	def __init__(self):
		self.Protocol = None
		self.Command  = None
		self.Status   = None
		self.Flags    = None
		self.Flags2   = None
		self.PIDHigh  = None
		self.SecurityFeatures = None
		self.Signature = None
		self.Reserved = None
		self.TID      = None
		self.PIDLow   = None
		self.UID      = None
		self.MID      = None

	@staticmethod
	def from_bytes(bbuff):
		return SMBHeader.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		hdr = SMBHeader()
		hdr.Protocol = buff.read(4)
		assert hdr.Protocol == b'\xFFSMB', "SMBv1 Header Magic incorrect!"
		hdr.Command  = SMBCommand(int.from_bytes(buff.read(1), byteorder='little', signed = False))
		hdr.Status   = NTStatus(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		hdr.Flags    = SMBHeaderFlagsEnum(int.from_bytes(buff.read(1), byteorder='little', signed = False))
		hdr.Flags2   = SMBHeaderFlags2Enum(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		hdr.PIDHigh  = int.from_bytes(buff.read(2), byteorder='little', signed = False)

		if SMBHeaderFlags2Enum.SMB_FLAGS2_SMB_SECURITY_SIGNATURE in hdr.Flags2:
			hdr.SecurityFeatures = buff.read(8)
		else:
			hdr.Signature = buff.read(8)

		hdr.Reserved = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		hdr.TID      = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		hdr.PIDLow   = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		hdr.UID      = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		hdr.MID      = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		return hdr

	@staticmethod
	def construct(command, status, flags, flags2, uid = 0, mid = 0, tid = 0, securityfeatures = None, signature = None, pidhigh = 0, pidlow = 0):
		hdr = SMBHeader()
		hdr.Protocol = b'\xFFSMB'
		hdr.Command  = command
		hdr.Status   = status
		hdr.Flags    = flags
		hdr.Flags2   = flags2
		hdr.PIDHigh  = pidhigh

		if SMBHeaderFlags2Enum.SMB_FLAGS2_SMB_SECURITY_SIGNATURE in hdr.Flags2:
			if securityfeatures is None:
				raise Exception('SMB_FLAGS2_SMB_SECURITY_SIGNATURE is present but SecurityFeatures was not supplied!')
			hdr.SecurityFeatures = securityfeatures
		else:
			if signature is not None:
				hdr.Signature = signature

		hdr.Reserved = 0
		hdr.TID      = tid
		hdr.PIDLow   = pidlow
		hdr.UID      = uid
		hdr.MID      = mid

		return hdr

	def to_bytes(self):
		t  = self.Protocol
		t += self.Command.value.to_bytes(1, byteorder = 'little', signed=False)
		t += self.Status.value.to_bytes(4, byteorder = 'little', signed=False)
		t += self.Flags.to_bytes(1, byteorder = 'little', signed=False)
		t += self.Flags2.value.to_bytes(2, byteorder = 'little', signed=False)
		t += self.PIDHigh.to_bytes(2, byteorder = 'little', signed=False)
		if self.SecurityFeatures is not None:
			t += self.SecurityFeatures
		elif self.Signature is not None:
			t += self.Signature
		else:
			t += b'\x00'*8
		t += self.Reserved.to_bytes(2, byteorder = 'little', signed=False)
		t += self.TID.to_bytes(2, byteorder = 'little', signed=False)
		t += self.PIDLow.to_bytes(2, byteorder = 'little', signed=False)
		t += self.UID.to_bytes(2, byteorder = 'little', signed=False)
		t += self.MID.to_bytes(2, byteorder = 'little', signed=False)
		return t

	def __repr__(self):
		t = '===SMBHeader===\r\n'
		t += 'Command: %s\r\n' % self.Command.name
		t += 'Flags:   %s\r\n' % repr(self.Flags)
		t += 'Flags2:  %s\r\n' % repr(self.Flags2)
		t += 'PIDHigh: %s\r\n' % self.PIDHigh
		t += 'SecurityFeatures: %s\r\n' % (self.SecurityFeatures.hex() if self.SecurityFeatures is not None else 'NONE')
		t += 'Reserved: %s\r\n' % self.Reserved
		t += 'TID: %s\r\n' % self.TID
		t += 'PIDLow: %s\r\n' % self.PIDLow
		t += 'UID: %s\r\n' % self.UID
		t += 'MID: %s\r\n' % self.MID
		return t


# https://msdn.microsoft.com/en-us/library/ee441946.aspx
class SMB_COM_NEGOTIATE_REPLY:
	def __init__(self):
		##### SMB_Parameters #####
		self.WordCount = None
		self.DialectIndex = None # this is for really really old protocol dialects
		self.SecurityMode  = None
		self.MaxMpxCount   = None
		self.MaxNumberVcs  = None
		self.MaxBufferSize = None
		self.MaxRawSize    = None
		self.SessionKey    = None
		self.Capabilities  = None
		self.SystemTime    = None
		self.ServerTimeZone = None
		self.ChallengeLength= None
		##### SMB_Data #####
		self.ByteCount  = None
		self.Challenge  = None
		self.DomainName = None

		self.uuid = None
		self.secblob = None

	@staticmethod
	def from_bytes(bbuff):
		return SMB_COM_NEGOTIATE_REPLY.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		msg = SMB_COM_NEGOTIATE_REPLY()
		msg.WordCount       = int.from_bytes(buff.read(1), byteorder='little', signed = False)
		msg.DialectIndex    = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.SecurityMode    = SMBSecurityMode(buff.read(1))
		msg.MaxMpxCount     = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.MaxNumberVcs    = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.MaxBufferSize   = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.MaxRawSize      = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.SessionKey      = buff.read(4)

		msg.Capabilities    = SMBCapabilities(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		msg.SystemTime      = wintime2datetime(int.from_bytes(buff.read(8), byteorder='little', signed = False))
		msg.ServerTimeZone  = int.from_bytes(buff.read(2), byteorder='little', signed = True)
		msg.ChallengeLength = int.from_bytes(buff.read(1), byteorder='little', signed = False)

		msg.ByteCount       = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.Challenge       = buff.read(msg.ChallengeLength)
		msg.DomainName      = buff.read().deocde()
		return msg

	@staticmethod
	def construct(dialectindex, securitymode, sessionkey, capabilities, uuid, secblob, systemtime = datetime.datetime.utcnow()):
		msg = SMB_COM_NEGOTIATE_REPLY()
		msg.WordCount       = 0x11
		msg.DialectIndex    = dialectindex
		msg.SecurityMode    = securitymode
		msg.MaxMpxCount     = 50
		msg.MaxNumberVcs    = 1
		msg.MaxBufferSize   = 16644
		msg.MaxRawSize      = 65536
		msg.SessionKey      = sessionkey

		msg.Capabilities    = capabilities
		msg.SystemTime      = systemtime
		msg.ServerTimeZone  = 0
		msg.ChallengeLength = 0

		msg.ByteCount       = len(uuid.bytes_le) + len(secblob)
		msg.uuid = uuid
		msg.secblob = secblob

		return msg

	def to_bytes(self):
		t  = b''
		t += self.WordCount.to_bytes(1, byteorder = 'little', signed=False)
		t += self.DialectIndex.to_bytes(2, byteorder = 'little', signed=False)
		t += self.SecurityMode.value.to_bytes(1, byteorder = 'little', signed=False)
		t += self.MaxMpxCount.to_bytes(2, byteorder = 'little', signed=False)
		t += self.MaxNumberVcs.to_bytes(2, byteorder = 'little', signed=False)
		t += self.MaxBufferSize.to_bytes(4, byteorder = 'little', signed=False)
		t += self.MaxRawSize.to_bytes(4, byteorder = 'little', signed=False)
		t += self.SessionKey
		t += self.Capabilities.to_bytes(4, byteorder = 'little', signed=False)
		t += dt2wt(self.SystemTime).to_bytes(8, byteorder = 'little', signed=False)
		t += self.ServerTimeZone.to_bytes(2, byteorder = 'little', signed=True)
		t += self.ChallengeLength.to_bytes(1, byteorder = 'little', signed=False)
		t += self.ByteCount.to_bytes(2, byteorder = 'little', signed=False)
		t += self.uuid.bytes_le
		t += self.secblob

		return t


	def __repr__(self):
		t = '=== SMB_COM_NEGOTIATE_REPLY ===\r\n'
		t += 'WordCount:     %d\r\n' % self.WordCount
		t += 'DialectIndex:  %d\r\n' % self.DialectIndex
		t += 'SecurityMode:  %s\r\n' % repr(self.SecurityMode)
		t += 'MaxMpxCount:   %d\r\n' % self.MaxMpxCount
		t += 'MaxNumberVcs:  %d\r\n' % self.MaxNumberVcs
		t += 'MaxBufferSize: %d\r\n' % self.MaxBufferSize
		t += 'MaxRawSize:    %d\r\n' % self.MaxRawSize
		t += 'SessionKey:    %s\r\n' % self.SessionKey.hex()
		t += 'Capabilities:  %s\r\n' % repr(self.Capabilities)
		t += 'SystemTime:    %s\r\n' % self.SystemTime.isoformat()
		t += 'ServerTimeZone:    %s\r\n' % self.ServerTimeZone
		t += 'ChallengeLength:    %d\r\n' % self.ChallengeLength

		return t


# https://msdn.microsoft.com/en-us/library/ee441913.aspx
class SMB_COM_NEGOTIATE_REQ:
	def __init__(self):
		##### parameters ####
		self.WordCount = None
		##### SMB_Data ###
		self.ByteCount = None
		self.Dialects  = None

	@staticmethod
	def from_bytes(bbuff):
		return SMB_COM_NEGOTIATE_REQ.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		cmd = SMB_COM_NEGOTIATE_REQ()
		cmd.WordCount = int.from_bytes(buff.read(1), byteorder='little', signed = False)
		if cmd.WordCount > 0:
			cmd.DialectIndex = []
			for i in range(cmd.WordCount):
				cmd.DialectIndex.append(int.from_bytes(buff.read(1), byteorder='little', signed = False))
		cmd.ByteCount = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		cmd.Dialects  = []
		for raw in buff.read(cmd.ByteCount).split(b'\x00'):
			if raw == b'':
				continue
			cmd.Dialects.append(SMB_Dialect(raw))
		return cmd

	def __repr__(self):
		t = '===SMB_COM_NEGOTIATE_REQ===\r\n'
		t += repr(self.params)
		t += repr(self.data)	
		return t


# https://msdn.microsoft.com/en-us/library/ee441849.aspx
class SMB_COM_SESSION_SETUP_ANDX_REQ:
	def __init__(self):
		##### parameters ####
		self.WordCount     = None
		self.AndXCommand   = None
		self.AndXReserved  = None
		self.AndXOffset    = None
		self.MaxBufferSize = None
		self.MaxMpxCount   = None
		self.VcNumber      = None
		self.SessionKey    = None
		self.SecurityBlobLen = None
		self.Reserved      = None
		self.Capabilities  = None
		##### SMB_Data ###
		self.ByteCount     = None
		self.SecurityBlob  = None
		self.NativeOS      = None
		self.NativeLanMan  = None

	@staticmethod
	def from_bytes(bbuff):
		return SMB_COM_SESSION_SETUP_ANDX_REQ.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		cmd = SMB_COM_SESSION_SETUP_ANDX_REQ()
		cmd.WordCount     = int.from_bytes(buff.read(1), byteorder='little', signed = False)
		cmd.AndXCommand   = int.from_bytes(buff.read(1), byteorder='little', signed = False)
		cmd.AndXReserved  = int.from_bytes(buff.read(1), byteorder='little', signed = False)
		cmd.AndXOffset    = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		cmd.MaxBufferSize = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		cmd.MaxMpxCount   = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		cmd.VcNumber      = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		cmd.SessionKey    = buff.read(4)
		cmd.SecurityBlobLen = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		cmd.Reserved      = buff.read(4)
		cmd.Capabilities  = SMBCapabilities(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		cmd.ByteCount     = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		cmd.SecurityBlob  = buff.read(cmd.SecurityBlobLen)
		#be careful of padding at this point, a padding byte of \x00 might be inserted to keep the strings on a 16 byte alignment
		pos = buff.tell()
		if pos %2 != 0:
			buff.read(1)

		#this parsing is disgusting, but i have no better idea...
		t =  buff.read()
		print(t)
		print(t.decode('utf-16-le'))
		print(t.split(b'\x00\x00\x00'))
		t1, t2, *aaa = t.split(b'\x00\x00\x00')
		cmd.NativeOS      = (t1+b'\x00').decode('utf-16-le')
		cmd.NativeLanMan  = (t2+b'\x00').decode('utf-16-le')

		return cmd

	def __repr__(self):
		t = '===SMB_COM_SESSION_SETUP_AND_X_REQ===\r\n'
		t += repr(self.params)
		t += repr(self.data)	
		return t

# https://msdn.microsoft.com/en-us/library/ee442143.aspx
class SMB_COM_SESSION_SETUP_ANDX_REPLY:
	def __init__(self):
		##### parameters ####
		self.WordCount     = None
		self.AndXCommand   = None
		self.AndXReserved  = None
		self.AndXOffset    = None
		self.Action        = None
		self.SecurityBlobLen = None
		##### SMB_Data ###
		self.ByteCount     = None
		self.SecurityBlob  = None
		self.NativeOS      = None
		self.NativeLanMan  = None

	@staticmethod
	def from_bytes(bbuff):
		return SMB_COM_SESSION_SETUP_ANDX_REPLY.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		cmd = SMB_COM_SESSION_SETUP_ANDX_REPLY()
		cmd.WordCount     = int.from_bytes(buff.read(1), byteorder='little', signed = False)
		cmd.AndXCommand   = int.from_bytes(buff.read(1), byteorder='little', signed = False)
		cmd.AndXReserved  = int.from_bytes(buff.read(1), byteorder='little', signed = False)
		cmd.AndXOffset    = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		cmd.Action        = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		cmd.SecurityBlobLen = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		cmd.ByteCount     = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		cmd.SecurityBlob  = buff.read(cmd.SecurityBlobLen)
		#be careful of padding at this point, a padding byte of \x00 might be inserted to keep the strings on a 16 byte alignment
		pos = buff.tell()
		if pos %2 != 0:
			buff.read(1)

		#this parsing is disgusting, but i have no better idea...
		t =  buff.read()
		t1, t2, *aaa = t.split(b'\x00\x00\x00')
		cmd.NativeOS      = t1.decode('utf16-le')
		cmd.NativeLanMan  = t2.decode('utf16-le')

		return cmd

	@staticmethod
	def construct(secblob = None, nativeos = None, nativelanman = None):
		cmd = SMB_COM_SESSION_SETUP_ANDX_REPLY()
		cmd.WordCount     = 4
		cmd.AndXCommand   = 0xff
		cmd.AndXReserved  = 0
		cmd.AndXOffset    = 0
		cmd.Action        = 0
		cmd.SecurityBlobLen = len(secblob)
		cmd.ByteCount     = None #to be set when to_bytes is invoked
		cmd.SecurityBlob  = secblob
		cmd.NativeOS      = nativeos
		cmd.NativeLanMan  = nativelanman
		return cmd

	def to_bytes(self):
		t  = self.WordCount.to_bytes(1, byteorder = 'little', signed=False)
		t += self.AndXCommand.to_bytes(1, byteorder = 'little', signed=False)
		t += self.AndXReserved.to_bytes(1, byteorder = 'little', signed=False)
		t += self.AndXOffset.to_bytes(2, byteorder = 'little', signed=False)
		t += self.Action.to_bytes(2, byteorder = 'little', signed=False)
		t += self.SecurityBlobLen.to_bytes(2, byteorder = 'little', signed=False)

		if self.NativeOS is None:
			nativeos = b''
		else:
			nativeos = (self.NativeOS + '\x00').encode('utf-16-le')
		
		if self.NativeLanMan is None:
			nativelanman = b''
		else:
			nativelanman = (self.NativeLanMan + '\x00').encode('utf-16-le')
		tlen = 2 + self.SecurityBlobLen + len(t)
		padneeded = tlen %2 != 0

		if padneeded:
			tlen += 1

		t += tlen.to_bytes(2, byteorder = 'little', signed=False) #this is he bytecount
		t += self.SecurityBlob
		if padneeded:
			t += b'\x00'
		t += nativeos
		t += nativelanman

		return t

	def __repr__(self):
		t = '===SMB_COM_SESSION_SETUP_ANDX_REPLY===\r\n'
		t += repr(self.params)
		t += repr(self.data)	
		return t


class SMBSetupAction(enum.Enum):
	SMB_SETUP_GUEST = 0x0001
	SMB_SETUP_USE_LANMAN_KEY = 0x0002


# https://msdn.microsoft.com/en-us/library/ee441616.aspx
class SMBCommand(enum.Enum): #SMB_COM
	SMB_COM_CREATE_DIRECTORY = 0x00
	SMB_COM_DELETE_DIRECTORY = 0x01
	SMB_COM_OPEN = 0x02
	SMB_COM_CREATE  = 0x03
	SMB_COM_CLOSE = 0x04
	SMB_COM_FLUSH = 0x05
	SMB_COM_DELETE = 0x06
	SMB_COM_RENAME = 0x07
	SMB_COM_QUERY_INFORMATION = 0x08
	SMB_COM_SET_INFORMATION = 0x09
	SMB_COM_READ = 0x0A
	SMB_COM_WRITE = 0x0B
	SMB_COM_LOCK_BYTE_RANGE = 0x0C
	SMB_COM_UNLOCK_BYTE_RANGE = 0x0D
	SMB_COM_CREATE_TEMPORARY = 0x0E
	SMB_COM_CREATE_NEW = 0x0F
	SMB_COM_CHECK_DIRECTORY = 0x10
	SMB_COM_PROCESS_EXIT = 0x11
	SMB_COM_SEEK = 0x12
	SMB_COM_LOCK_AND_READ = 0x13
	SMB_COM_WRITE_AND_UNLOCK = 0x14
	#Unused 0x15-0x19
	SMB_COM_READ_RAW = 0x1A
	SMB_COM_READ_MPX = 0x1B
	SMB_COM_READ_MPX_SECONDARY = 0x1C
	SMB_COM_WRITE_RAW = 0x1D
	SMB_COM_WRITE_MPX = 0x1E
	SMB_COM_WRITE_MPX_SECONDARY = 0x1F
	SMB_COM_WRITE_COMPLETE = 0x20
	SMB_COM_QUERY_SERVER  = 0x21
	SMB_COM_SET_INFORMATION2 = 0x22
	##### TODODODODODODODO!!!!!!!!!!!!!!!!!
	SMB_COM_NEGOTIATE = 0x72
	SMB_COM_SESSION_SETUP_ANDX = 0x73
	SMB_COM_LOGOFF_ANDX = 0x74
	SMB_COM_TREE_CONNECT_ANDX = 0x75


class SMBHeaderFlagsEnum(enum.IntFlag):
	SMB_FLAGS_LOCK_AND_READ_OK = 0x01
	SMB_FLAGS_BUF_AVAIL = 0x02
	Reserved = 0x04
	SMB_FLAGS_CASE_INSENSITIVE = 0x08
	SMB_FLAGS_CANONICALIZED_PATHS = 0x10
	SMB_FLAGS_OPLOCK = 0x20
	SMB_FLAGS_OPBATCH = 0x40
	SMB_FLAGS_REPLY = 0x80


class SMBHeaderFlags2Enum(enum.IntFlag):
	SMB_FLAGS2_LONG_NAMES = 0x0001
	SMB_FLAGS2_EAS = 0x0002
	SMB_FLAGS2_SMB_SECURITY_SIGNATURE = 0x0004
	SMB_FLAGS2_IS_LONG_NAME = 0x0040
	SMB_FLAGS2_EXTENDED_SECURITY = 0x0800
	SMB_FLAGS2_DFS = 0x1000
	SMB_FLAGS2_PAGING_IO = 0x2000
	SMB_FLAGS2_NT_STATUS = 0x4000
	SMB_FLAGS2_UNICODE = 0x8000


class SMBSecurityMode(enum.IntFlag):
	NEGOTIATE_USER_SECURITY = 0x01
	NEGOTIATE_ENCRYPT_PASSWORDS = 0x02
	NEGOTIATE_SECURITY_SIGNATURES_ENABLED = 0x04
	NEGOTIATE_SECURITY_SIGNATURES_REQUIRED = 0x08
	#others are Reserved


class SMBCapabilities(enum.IntFlag):
	CAP_RAW_MODE         = 0x00000001
	CAP_MPX_MODE         = 0x00000002
	CAP_UNICODE          = 0x00000004
	CAP_LARGE_FILES      = 0x00000008
	CAP_NT_SMBS          = 0x00000010
	CAP_RPC_REMOTE_APIS  = 0x00000020
	CAP_STATUS32         = 0x00000040
	CAP_LEVEL_II_OPLOCKS = 0x00000080
	CAP_LOCK_AND_READ    = 0x00000100
	CAP_NT_FIND          = 0x00000200
	CAP_BULK_TRANSFER    = 0x00000400
	CAP_COMPRESSED_DATA  = 0x00000800
	CAP_DFS              = 0x00001000
	CAP_QUADWORD_ALIGNED = 0x00002000
	CAP_LARGE_READX      = 0x00004000
	CAP_NT_EXTENDED_SECURITY = 0x80000000


class SMB_Dialect:
	def __init__(self, data = None):
		self.BufferFormat = None
		self.DialectString = None

		if data is not None:
			self.parse(data)

	""" TODO!!!
	def from_bytes(bbuff):
		return SMB_Dialect.from_buffer(io.BytesIO(bbuff))

	def from_buffer(buff):
		cmd = SMB_Dialect()
		cmd.BufferFormat  = buff.read(1)
		cmd.DialectString = SMB_COM_NEGOTIATE_REQ_DATA.from_buffer(buff)
		return cmd
	"""
	def parse(self, data):
		self.BufferFormat  = data[0]
		self.DialectString = data[1:].decode('ascii')

	def __repr__(self):
		t = ''
		t += 'DialectString: %s' % self.DialectString
		return t

class SMBMessage:
	def __init__(self):
		self.type      = 1
		self.header    = None
		self.command   = None

	@staticmethod
	def from_bytes(bbuff):
		return SMBMessage.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		msg = SMBMessage()
		msg.header = SMBHeader.from_buffer(buff)
		classname = msg.header.Command.name
		if SMBHeaderFlagsEnum.SMB_FLAGS_REPLY in msg.header.Flags:
			classname += '_REPLY'
		else:
			classname += '_REQ'
		class_ = getattr(sys.modules[__name__], classname)
		msg.command = class_.from_buffer(buff)
		
		return msg

	def to_bytes(self):
		t  = self.header.to_bytes()
		t += self.command.to_bytes() 
		return t

	def __repr__(self):
		t = '== SMBMessage ==\r\n'
		t += repr(self.header)
		t += repr(self.command)
		return t




# https://blog.pythian.com/repost-oracle-protocol/
# https://github.com/SpiderLabs/net-tns
# http://ckng62.blogspot.com/2014/02/tns-data-packet-structure.html
# http://www.nyoug.org/Presentations/2008/Sep/Harris_Listening%20In.pdf

import enum
import io

from responder3.core.asyncio_helpers import *

class TNSPacketType(enum.Enum):
	UNK_0 = 0
	CONNECT = 1
	ACCEPT = 2
	ACK = 3
	REFUSE = 4
	REDIRECT = 5
	DATA = 6
	NULL = 7
	UNK_8 = 8
	ABORT = 9
	UNK_10 = 10
	RESEND = 11
	MARKER = 12
	ATTENTION = 13
	CONTROL = 14
	UNK_15 = 15

class TNSPacket:
	def __init__(self):
		self.header = None
		self.payload = None

	def to_bytes(self):
		data = self.payload.to_bytes()
		if not self.header:
			self.header = TNSHeader()
			self.header.packet_length = len(data) + 8
			self.header.packet_checksum = 0
			self.header.packet_type = tnstype2class_inv[type(self.payload)]
			self.header.flags = 0
			self.header.checksum = 0		
		
		return self.header.to_bytes() + data

	@staticmethod
	def from_bytes(bbuff):
		return TNSPacket.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		packet = TNSPacket() 
		packet.header = TNSHeader.from_buffer(buff)
		packet.payload = tnstype2class[packet.header.packet_type].from_buffer(buff)
		return packet

	@staticmethod
	async def from_streamreader(reader):
		t_length = await readexactly_or_exc(reader, 2)
		length = int.from_bytes(t_length,byteorder = 'big', signed = False) - 2
		data = await readexactly_or_exc(reader, length)
		return TNSPacket.from_bytes(t_length + data)


class TNSHeader:
	def __init__(self):
		self.packet_length = None
		self.packet_checksum = None
		self.packet_type = None
		self.flags = None
		self.checksum = None

	@staticmethod
	def from_bytes(bbuff):
		return TNSHeader.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		hdr = TNSHeader() 
		hdr.packet_length = int.from_bytes(buff.read(2),byteorder = 'big', signed = False)
		hdr.packet_checksum = int.from_bytes(buff.read(2),byteorder = 'big', signed = False)
		hdr.packet_type = TNSPacketType(int.from_bytes(buff.read(1),byteorder = 'big', signed = False))
		hdr.flags = int.from_bytes(buff.read(1),byteorder = 'big', signed = False)
		hdr.checksum = int.from_bytes(buff.read(2),byteorder = 'big', signed = False)
		return hdr

	def to_bytes(self):
		t  = self.packet_length.to_bytes(2, byteorder = 'big', signed = False)
		t += self.packet_checksum.to_bytes(2, byteorder = 'big', signed = False)
		t += self.packet_type.value.to_bytes(1, byteorder = 'big', signed = False)
		t += self.flags.to_bytes(1, byteorder = 'big', signed = False)
		t += self.checksum.to_bytes(2, byteorder = 'big', signed = False)
		return t

class TNSConnect:
	def __init__(self):
		self.maximum_version = None
		self.minimum_version = None
		self.service_flags = None
		self.sdu_size = None
		self.maximum_tdu_size = None
		self.protocol_flags = None
		self.line_turnaround_value = None
		self.byte_order = None
		self.data_length = None
		self.data_offset = None
		self.maximum_connect_receive = None
		self.flags1 = None
		self.flags2 = None
		self.trace_item1 = None
		self.trace_item2 = None
		self.trace_connection_id = None
		self.unknown = None
		self.data = None

	@staticmethod
	def from_bytes(bbuff):
		return TNSConnect.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		tns = TNSConnect() 
		tns.maximum_version = int.from_bytes(buff.read(2),byteorder = 'big', signed = False)
		tns.minimum_version = int.from_bytes(buff.read(2),byteorder = 'big', signed = False)
		tns.service_flags = int.from_bytes(buff.read(2),byteorder = 'big', signed = False)
		tns.sdu_size = int.from_bytes(buff.read(2),byteorder = 'big', signed = False)
		tns.maximum_tdu_size = int.from_bytes(buff.read(2),byteorder = 'big', signed = False)
		tns.protocol_flags = int.from_bytes(buff.read(2),byteorder = 'big', signed = False)
		tns.line_turnaround_value = int.from_bytes(buff.read(2),byteorder = 'big', signed = False)
		tns.byte_order = int.from_bytes(buff.read(2),byteorder = 'big', signed = False)
		tns.data_length = int.from_bytes(buff.read(2),byteorder = 'big', signed = False)
		tns.data_offset = int.from_bytes(buff.read(2),byteorder = 'big', signed = False)
		tns.maximum_connect_receive = int.from_bytes(buff.read(2),byteorder = 'big', signed = False)
		tns.flags1 = int.from_bytes(buff.read(1),byteorder = 'big', signed = False)
		tns.flags2 = int.from_bytes(buff.read(1),byteorder = 'big', signed = False)
		tns.trace_item1 = int.from_bytes(buff.read(4),byteorder = 'big', signed = False)
		tns.trace_item2 = int.from_bytes(buff.read(4),byteorder = 'big', signed = False)
		tns.trace_connection_id = int.from_bytes(buff.read(8),byteorder = 'big', signed = False)
		tns.unknown = int.from_bytes(buff.read(8),byteorder = 'big', signed = False)
		tns.data = buff.read(-1).decode()
		return tns

class TNSAccept:
	def __init__(self):
		self.version = None
		self.service_flags = None
		self.sdu_size = None
		self.maximum_tdu_size = None
		self.byte_order = None
		self.data_length = None
		self.data_offset = None
		self.flags1 = None
		self.flags2 = None
		self.padding = None
		self.data = None

	@staticmethod
	def from_bytes(bbuff):
		return TNSAccept.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		tns = TNSAccept()
		tns.version = int.from_bytes(buff.read(2),byteorder = 'big', signed = False)
		tns.service_flags = int.from_bytes(buff.read(2),byteorder = 'big', signed = False)
		tns.sdu_size = int.from_bytes(buff.read(2),byteorder = 'big', signed = False)
		tns.maximum_tdu_size = int.from_bytes(buff.read(2),byteorder = 'big', signed = False)
		tns.byte_order = int.from_bytes(buff.read(2),byteorder = 'big', signed = False)
		tns.data_length = int.from_bytes(buff.read(2),byteorder = 'big', signed = False)
		tns.data_offset = int.from_bytes(buff.read(2),byteorder = 'big', signed = False)
		tns.flags1 = int.from_bytes(buff.read(1),byteorder = 'big', signed = False)
		tns.flags2 = int.from_bytes(buff.read(1),byteorder = 'big', signed = False)
		tns.padding = buff.read(tns.data_offset - 24)
		tns.data = buff.read(tns.data_length)

		return tns

	def to_bytes(self):
		t = self.version.to_bytes(2, byteorder = 'big', signed = False)
		t += self.service_flags.to_bytes(2, byteorder = 'big', signed = False)
		t += self.sdu_size.to_bytes(2, byteorder = 'big', signed = False)
		t += self.maximum_tdu_size.to_bytes(2, byteorder = 'big', signed = False)
		t += self.byte_order.to_bytes(2, byteorder = 'big', signed = False)
		self.data_length = len(self.data) if self.data else 0
		t += self.data_length.to_bytes(2, byteorder = 'big', signed = False)
		t += self.data_offset.to_bytes(2, byteorder = 'big', signed = False)
		t += self.flags1.to_bytes(1, byteorder = 'big', signed = False)
		t += self.flags2.to_bytes(1, byteorder = 'big', signed = False)
		if self.padding:
			t += self.padding
		if self.data:
			t += self.data
		return t

class TNSRefuse:
	def __init__(self):
		self.user_reason = None
		self.system_reason = None
		self.data_length = None
		self.data = None

	@staticmethod
	def from_bytes(bbuff):
		return TNSRefuse.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		tns = TNSRefuse()
		tns.user_reason = int.from_bytes(buff.read(1),byteorder = 'big', signed = False)
		tns.system_reason = int.from_bytes(buff.read(1),byteorder = 'big', signed = False)
		tns.data_length = int.from_bytes(buff.read(2),byteorder = 'big', signed = False)
		tns.data = buff.read(tns.data_length)
		
		return tns

class TNSRedirect:
	def __init__(self):
		self.data_length = None
		self.data = None

	@staticmethod
	def from_bytes(bbuff):
		return TNSRedirect.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		tns = TNSRedirect()
		tns.data_length = int.from_bytes(buff.read(2),byteorder = 'big', signed = False)
		tns.data = buff.read(tns.data_length)
		return tns

class TNSResend:
	def __init__(self):
		pass

	def to_bytes(self):
		return b''

class TNSNull:
	def __init__(self):
		pass

class TNSAck:
	def __init__(self):
		pass

class TNSAttention:
	def __init__(self):
		pass

class TNSControl:
	def __init__(self):
		pass

class TNSAbort:
	def __init__(self):
		self.user_reason = None
		self.system_reason = None
		self.data = None

	@staticmethod
	def from_bytes(bbuff):
		return TNSAbort.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		tns = TNSAbort()
		tns.user_reason = int.from_bytes(buff.read(1),byteorder = 'big', signed = False)
		tns.system_reason = int.from_bytes(buff.read(1),byteorder = 'big', signed = False)
		tns.data = buff.read(-1)
		return tns

class TNSData:
	def __init__(self):
		self.flags = None
		self.data = None

	@staticmethod
	def from_bytes(bbuff):
		return TNSData.from_buffer(io.BytesIO(bbuff))

	# TODO: more processing of data
	@staticmethod
	def from_buffer(buff):
		tns = TNSData()
		tns.flags = int.from_bytes(buff.read(2),byteorder = 'big', signed = False)
		tns.data = buff.read(-1)

		return tns

	def to_bytes(self):
		t = self.flags.to_bytes(2 ,byteorder = 'big', signed = False)
		t += self.data
		return t

class TNSMarker:
	def __init__(self):
		self.marker_type = None
		self.data = None

	@staticmethod
	def from_bytes(bbuff):
		return TNSMarker.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		tns = TNSMarker()
		tns.marker_type = int.from_bytes(buff.read(1),byteorder = 'big', signed = False)
		tns.data = buff.read(-1)

		return tns

class TNSXXX:
	def __init__(self):
		self.data = None

	@staticmethod
	def from_bytes(bbuff):
		return TNSXXX.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		tns = TNSXXX()
		tns.data = buff.read(-1)
		return tns

tnstype2class = {
	TNSPacketType.UNK_0 : TNSXXX,
	TNSPacketType.CONNECT : TNSConnect,
	TNSPacketType.ACCEPT : TNSAccept,
	TNSPacketType.ACK : TNSAck,
	TNSPacketType.REFUSE : TNSRefuse,
	TNSPacketType.REDIRECT : TNSRedirect,
	TNSPacketType.DATA : TNSData,
	TNSPacketType.NULL : TNSNull,
	TNSPacketType.UNK_8 : TNSXXX,
	TNSPacketType.ABORT : TNSAbort,
	TNSPacketType.UNK_10 : TNSXXX,
	TNSPacketType.RESEND : TNSResend,
	TNSPacketType.MARKER : TNSMarker,
	TNSPacketType.ATTENTION : TNSAttention,
	TNSPacketType.CONTROL : TNSControl,
	TNSPacketType.UNK_15 : TNSXXX,
}

tnstype2class_inv = {v: k for k, v in tnstype2class.items()}


# https://tools.ietf.org/html/rfc5321
# https://stackoverflow.com/questions/8022530/python-check-for-valid-email-address
import re
import io
import enum
import asyncio
import ipaddress
import traceback
from base64 import b64decode, b64encode

from responder3.core.commons import read_element
from responder3.core.logging.log_objects import Credential
from responder3.core.asyncio_helpers import *

EMAIL_REGEX = re.compile(r"[^@]+@[^@]+\.[^@]+")


class SMTPServerState(enum.Enum):
	START           = enum.auto()
	DATAINCOMING    = enum.auto()
	DATAFINISHED    = enum.auto()
	NOTAUTHETICATED = enum.auto()
	AUTHSTARTED     = enum.auto()
	AUTHENTICATED   = enum.auto()
	MAIL            = enum.auto()


SMTPReplyCode = {
	211: 'System status, or system help reply',
	214: 'Help message', #(Information on how to use the receiver or the meaning of a particular non-standard command; this reply is useful only to the human user)
	220: '{domain} Service ready',
	221: '{domain} Service closing transmission channel',
	235: 'Authentication Succeeded',
	250: 'Requested mail action okay, completed',
	251: 'User not local; will forward to <forward-path>', #(See Section 3.4)
	252: 'Cannot VRFY user, but will accept message and attempt delivery', #(See Section 3.5.3)
	334: '',
	354: 'Start mail input; end with <CRLF>.<CRLF>',
	421: '{domain} Service not available, closing transmission channel', #(This may be a reply to any command if the service knows it must shut down)
	432: 'A password transition is needed',
	450: 'Requested mail action not taken: mailbox unavailable', #(e.g., mailbox busy or temporarily blocked for policy reasons)
	451: 'Requested action aborted: local error in processing',
	452: 'Requested action not taken: insufficient system storage',
	454: 'Temporary authentication failure',
	455: 'Server unable to accommodate parameters',
	500: 'Syntax error, command unrecognized', #(This may include errors such as command line too long)
	501: 'Syntax error in parameters or arguments',
	502: 'Command not implemented', #(see Section 4.2.4)
	503: 'Bad sequence of commands',
	504: 'Command parameter not implemented',
	530: 'Authentication required',
	534: 'Authentication mechanism is too weak',
	535: 'Authentication credentials invalid',
	538: 'Encryption required for requested authentication  mechanism',
	550: 'Requested action not taken: mailbox unavailable', #(e.g., mailbox not found, no access, or command rejected for policy reasons)
	551: 'User not local; please try <forward-path>', #(See Section 3.4)
	552: 'Requested mail action aborted: exceeded storage allocation',
	553: 'Requested action not taken: mailbox name not allowed #(e.g., mailbox syntax incorrect)',
	554: 'Transaction failed', #(Or, in the case of a connection-opening response, "No SMTP service here")
	555: 'MAIL FROM/RCPT TO parameters not recognized or not implemented',
	666: 'A thousand nights we\'ve been calling your name Close your eyes but I won\'t go away We\'re there for you'
}


class SMTPCommand(enum.Enum):
	HELO = enum.auto()
	EHLO = enum.auto()
	MAIL = enum.auto()
	RCPT = enum.auto()
	DATA = enum.auto()
	RSET = enum.auto()
	VRFY = enum.auto()
	EXPN = enum.auto()
	HELP = enum.auto()
	NOOP = enum.auto()
	QUIT = enum.auto()
	AUTH = enum.auto()
	STARTTLS = enum.auto()
	XXXX = enum.auto() #this is a shortcut for unparsable input


SMTPMultilineCMD = [
	SMTPCommand.DATA,
]


class SMTPCommandParser:
	def __init__(self, encoding='ascii', timeout = 60):
		self.encoding = encoding
		self.timeout = timeout
		self.is_mulitline = False
		self.multiline_cmd = None
		self.multiline_buffer = None

	async def from_streamreader(self, reader):
		if self.is_mulitline:
			while True:
				temp = await readline_or_exc(reader, timeout=self.timeout)
				self.multiline_buffer += temp
				if temp == b'.\r\n':
					self.is_mulitline = False
					cmd = self.from_bytes(self.multiline_buffer)
					self.multiline_buffer = None
					return cmd

		buff = await readline_or_exc(reader, timeout=self.timeout)

		command, *params = buff.strip().decode(self.encoding).upper().split(' ')
		if command in SMTPCommand.__members__:
			if SMTPCommand[command] in SMTPMultilineCMD:
				self.is_mulitline = True
				self.multiline_buffer = buff
				return SMTPCMD[SMTPCommand[command]].from_bytes(None)

		return self.from_bytes(buff)

	def from_bytes(self, bbuff):
		return self.from_buffer(io.BytesIO(bbuff))

	def from_buffer(self, buff):
		pos = buff.tell()
		line = buff.readline()
		try:
			command, *params = line.strip().decode(self.encoding).upper().split(' ')
			if command in SMTPCommand.__members__:
				if SMTPCommand[command] in SMTPMultilineCMD:
					buff.seek(pos)
					return SMTPCMD[SMTPCommand[command]].from_bytes(buff.read())
				else:
					return SMTPCMD[SMTPCommand[command]].from_bytes(line)
			else:
				return SMTPXXXXCmd.from_bytes(line)

		except Exception as e:
			traceback.print_exc()
			return SMTPXXXXCmd.from_bytes(line)


# class SMTPHELOorEHLOCommand:
class SMTPHELOCmd:
	def __init__(self, encoding = 'ascii'):
		self.encoding = encoding
		self.command = SMTPCommand.HELO
		self.domain = None

	@staticmethod
	def from_buffer(buff, encoding='ascii'):
		return SMTPHELOCmd.from_bytes(buff.readline(), encoding)

	@staticmethod
	def from_bytes(bbuff, encoding='ascii'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = SMTPHELOCmd()
		t, bbuff = read_element(bbuff, toend=True)
		cmd.command = SMTPCommand[t.upper()]
		cmd.msgno, bbuff = read_element(bbuff, toend=True)
		return cmd

	@staticmethod
	def construct(domain):
		cmd = SMTPHELOCmd()
		cmd.domain = domain
		return cmd

	def to_bytes(self):
		return ('%s %s\r\n' % (self.command.value, self.domain)).encode(self.encoding)

	def __repr__(self):
		t = '== SMTP %s Command ==\r\n' % self.command.name
		t += 'Command : %s\r\n' % self.command.name
		t += 'Domain  : %s\r\n' % self.domain
		return t


class SMTPEHLOCmd:
	def __init__(self, encoding= 'ascii'):
		self.encoding = encoding
		self.command = SMTPCommand.EHLO
		self.domain = None

	@staticmethod
	def from_buffer(buff, encoding='ascii'):
		return SMTPEHLOCmd.from_bytes(buff.readline(), encoding)

	@staticmethod
	def from_bytes(bbuff, encoding='ascii'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = SMTPEHLOCmd()
		t, bbuff = read_element(bbuff, toend=True)
		cmd.command = SMTPCommand[t.upper()]
		cmd.msgno, bbuff = read_element(bbuff, toend=True)
		return cmd

	@staticmethod
	def construct(domain):
		cmd = SMTPEHLOCmd()
		cmd.domain = domain
		return cmd

	def to_bytes(self):
		return ('%s %s\r\n' % (self.command.value, self.domain)).encode(self.encoding)

	def __repr__(self):
		t = '== SMTP %s Command ==\r\n' % self.command.name
		t += 'Command : %s\r\n' % self.command.name
		t += 'Domain  : %s\r\n' % self.domain
		return t


class SMTPMAILCmd:
	def __init__(self, encoding= 'ascii'):
		self.encoding = encoding
		self.command = SMTPCommand.MAIL
		self.emailaddress = None
		self.params = None

	@staticmethod
	def from_buffer(buff, encoding='ascii'):
		return SMTPMAILCmd.from_bytes(buff.readline(), encoding)

	@staticmethod
	def from_bytes(bbuff, encoding='ascii'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = SMTPMAILCmd()
		t, bbuff = read_element(bbuff)
		cmd.command = SMTPCommand[t.upper()]
		t, bbuff = read_element(bbuff, marker=':')
		temp_email = bbuff.strip()
		if temp_email[0] == '<':
			temp_email = temp_email[1:]
			cmd.emailaddress, bbuff = read_element(temp_email, marker='>', toend= True)

		else:
			# this case means the email address was defined in an old format, using spaces
			cmd.emailaddress, bbuff = read_element(temp_email, toend= True)

		if bbuff != '':
			cmd.params = bbuff.split(' ')

		return cmd

	@staticmethod
	def construct(emailaddress, params = None):
		cmd = SMTPMAILCmd()
		cmd.emailaddress = emailaddress
		cmd.params = params
		return cmd

	def to_bytes(self):
		if self.params is not None:
			return ('%s %s %s\r\n' % (self.command.value, self.emailaddress, self.params)).encode(self.encoding)
		else:
			return ('%s %s\r\n' % (self.command.value, self.emailaddress)).encode(self.encoding)

	def __repr__(self):
		t = '== SMTP %s Command ==\r\n' % self.command.name
		t += 'Command : %s\r\n' % self.command.name
		t += 'emailaddress  : %s\r\n' % self.emailaddress
		t += 'params  : %s\r\n' % self.params
		return t


class SMTPRCPTCmd:
	# TODO: envelope command parsing?
	def __init__(self, encoding= 'ascii'):
		self.encoding = encoding
		self.command = SMTPCommand.RCPT
		self.emailaddress = None
		self.params = None

	@staticmethod
	def from_buffer(buff, encoding='ascii'):
		return SMTPRCPTCmd.from_bytes(buff.readline(), encoding)

	@staticmethod
	def from_bytes(bbuff, encoding='ascii'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = SMTPRCPTCmd()
		t, bbuff = read_element(bbuff)
		cmd.command = SMTPCommand[t.upper()]
		t, bbuff = read_element(bbuff, marker=':')
		temp_email = bbuff.strip()
		if temp_email[0] == '<':
			temp_email = temp_email[1:]
			cmd.emailaddress, bbuff = read_element(temp_email, marker='>', toend=True)

		else:
			# this case means the email address was defined in an old format, using spaces
			cmd.emailaddress, bbuff = read_element(temp_email, toend=True)

		if bbuff != '':
			cmd.params = bbuff.split(' ')

		return cmd

	@staticmethod
	def construct(emailaddress, params = None):
		cmd = SMTPRCPTCmd()
		cmd.emailaddress = emailaddress
		cmd.params = params
		return cmd

	def to_bytes(self):
		if self.params is not None:
			return ('%s %s %s\r\n' % (self.command.value, self.emailaddress, self.params)).encode(self.encoding)
		else:
			return ('%s %s\r\n' % (self.command.value, self.emailaddress)).encode(self.encoding)

	def __repr__(self):
		t = '== SMTP %s Command ==\r\n' % self.command.name
		t += 'Command : %s\r\n' % self.command.name
		t += 'emailaddress  : %s\r\n' % self.emailaddress
		t += 'params  : %s\r\n' % self.params
		return t


class SMTPResponseParser:
	def __init__(self, encoding='ascii', timeout = 60, multiline_buffer_size_limit = 4096):
		self.encoding = encoding
		self.timeout = timeout
		self.multiline_buffer_size_limit = multiline_buffer_size_limit

	async def from_streamreader(self, reader):
		is_multiline = True
		multiline_buffer = b''

		while True:
			buff = await readline_or_exc(reader, timeout=self.timeout)
			line = buff.strip().decode(self.encoding)
			if line[3] == '-':
				is_multiline = True
				multiline_buffer += buff
				if len(multiline_buffer) > self.multiline_buffer_size_limit:
					raise Exception('Multiline size limit reached!')
				continue

			else:
				if not is_multiline:
					return self.from_bytes(buff)
				else:
					multiline_buffer += buff
					if len(multiline_buffer) > self.multiline_buffer_size_limit:
						raise Exception('Multiline size limit reached!')
					return self.from_bytes(multiline_buffer)
				break

	def from_socket(self, sock):
		buffer = b''
		while True:
			temp = sock.recv(4096)
			if temp == b'':
				break

			buffer += temp
			bsize = len(buffer)
			if bsize > self.multiline_buffer_size_limit:
				raise Exception('Multiline size limit reached!')
			if bsize > 4:
				if buffer[:-1] == b'\n':
					if buffer[3] == b'-':
						continue
					else:
						return self.from_bytes(buffer)
				else:
					continue

	def from_bytes(self, bbuff):
		return self.from_buffer(io.BytesIO(bbuff))

	def from_buffer(self, buff):
		response = None
		lines = buff.readlines()
		for line in lines:
			try:
				temp = line.strip().decode(self.encoding)
				status_code = int(temp[:3], 10)
				if status_code not in SMTPReplyCode:
					raise Exception('Unknown REPLY code!')

				if response is None:
					response = SMTPReply()
					response.code = status_code
					response.parameter.append(temp[4:])
				else:
					if status_code != response.code:
						raise Exception('SMTP Multiline response with mismatching status codes!')

					response.parameter.append(temp[4:])

			except Exception as e:
				raise e

		return response


class SMTPReply:
	def __init__(self):
		self.code      = None
		self.parameter = []

	@staticmethod
	def construct(code, data = None):
		rep = SMTPReply()
		rep.code = code
		if data is None:
			rep.parameter = [SMTPReplyCode[code]]
		elif isinstance(data, str):
			rep.parameter = [data]
		elif isinstance(data, list):
			rep.parameter = data
		else:
			raise Exception('Unknown data for SMTP reply!')
		return rep

	def to_bytes(self):
		"""
		returns a list of bytes
		"""
		if self.parameter is None:
			raise Exception('Empty data for SMTP reply!')
		if len(self.parameter) == 1:
			return str(self.code).encode('ascii') + b' ' + self.parameter[-1].encode('ascii') + b'\r\n'
		else:
			temp = (str(self.code).encode('ascii') + b'-').join([b''] + [param.encode('ascii')+ b'\r\n' for param in self.parameter[:-1]])
			temp += str(self.code).encode('ascii') + b' ' + self.parameter[-1].encode('ascii') + b'\r\n'
			return temp
	

def checkEmailAddress(emailAddress):
	if not EMAIL_REGEX.match(emailAddress):
		raise Exception('Email FROM not an email address!')
	return True


class SMTPXXXXCmd:
	"""
	this is a container for unknown messages
	"""
	def __init__(self):
		self.command   = SMTPCommand.XXXX
		self.data  = None

	def construct(self):
		"""
		this should not be used for constructing
		"""
		pass

	@staticmethod
	def from_bytes(bbuff):
		return SMTPXXXXCmd.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		xxx = SMTPXXXXCmd()
		xxx.data = buff.read()
		return xxx

	def to_bytes(self):
		return self.data.encode('ascii')

	def __repr__(self):
		t = '== SMTP %s Command ==\r\n' % self.command.name
		t += 'data      : %s\r\n' % repr(self.data)
		return t

class SMTPSTARTTLSCmd:
	def __init__(self):
		self.command   = SMTPCommand.STARTTLS

	@staticmethod
	def from_bytes(bbuff, encoding='ascii'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = SMTPSTARTTLSCmd()
		return cmd

	def __repr__(self):
		t = '== SMTP %s Command ==\r\n' % self.command.name
		return t


class SMTPAUTHCmd:
	def __init__(self):
		self.command   = SMTPCommand.AUTH
		self.mechanism = None
		self.data  = None

	def construct(self, mechanism, data = None):
		self.mechanism = mechanism
		self.data  = None

	def to_bytes(self):
		if self.data is not None:
			return b' '.join([self.command.name.encode('ascii'), self.mechanism.encode('ascii'), self.data.encode('ascii')]) +b'\r\n'
		else:
			return b' '.join([self.command.name.encode('ascii'), self.mechanism.encode('ascii')]) +b'\r\n'

	@staticmethod
	def from_bytes(bbuff, encoding='ascii'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = SMTPAUTHCmd()
		t, bbuff = read_element(bbuff)
		cmd.command = SMTPCommand[t.upper()]
		cmd.mechanism, bbuff = read_element(bbuff, toend = True)
		if len(bbuff) > 0:
			cmd.data = bbuff
		return cmd

	def __repr__(self):
		t = '== SMTP %s Command ==\r\n' % self.command.name
		t += 'mechanism : %s\r\n' % repr(self.mechanism)
		t += 'data      : %s\r\n' % repr(self.data)
		return t


class SMTPQUITCmd:
	def __init__(self):
		self.command   = SMTPCommand.QUIT

	def construct(self):
		pass

	def to_bytes(self):
		return self.command.name.encode('ascii') + b'\r\n'

	@staticmethod
	def from_bytes(bbuff, encoding='ascii'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = SMTPQUITCmd()
		t, bbuff = read_element(bbuff, toend= True)
		cmd.command = SMTPCommand[t.upper()]
		return cmd

	def __repr__(self):
		t = '== SMTP %s Command ==\r\n' % self.command.name
		return t


class SMTPNOOPCmd:
	def __init__(self):
		self.command   = SMTPCommand.NOOP
		self.data      = None

	def construct(self, data):
		self.data      = data

	def to_bytes(self):
		if self.data is not None:
			return self.command.name.encode('ascii') + b' ' + ' '.join(self.data).encode('ascii') + b'\r\n'
		else:
			return self.command.name.encode('ascii') + b'\r\n'

	@staticmethod
	def from_bytes(bbuff, encoding='ascii'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = SMTPNOOPCmd()
		t, bbuff = read_element(bbuff, toend= True)
		cmd.command = SMTPCommand[t.upper()]
		if len(bbuff) > 0:
			cmd.data = bbuff.split(' ')
		return cmd

	def __repr__(self):
		t = '== SMTP %s Command ==\r\n' % self.command.name
		t += 'data  : %s\r\n' % repr(self.data)
		return t


class SMTPHELPCmd:
	def __init__(self):
		self.command   = SMTPCommand.HELP
		self.data      = None

	def construct(self, data):
		self.data      = data

	def to_bytes(self):
		if self.data is not None:
			return self.command.name.encode('ascii') + b' ' + ' '.join(self.data).encode('ascii') +b'\r\n'
		else:
			return self.command.name.encode('ascii') +b'\r\n'

	@staticmethod
	def from_bytes(bbuff, encoding='ascii'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = SMTPHELPCmd()
		t, bbuff = read_element(bbuff, toend= True)
		cmd.command = SMTPCommand[t.upper()]
		if len(bbuff) > 0:
			cmd.data = bbuff.split(' ')
		return cmd

	def __repr__(self):
		t = '== SMTP %s Command ==\r\n' % self.command.name
		t += 'data  : %s\r\n' % repr(self.data)
		return t


class SMTPEXPNCmd:
	def __init__(self):
		self.command   = SMTPCommand.VRFY
		self.username      = None

	def construct(self, data):
		self.username      = data

	def to_bytes(self):
		return self.command.name.encode('ascii') + b' ' +self.username.encode('ascii') +b'\r\n'

	@staticmethod
	def from_bytes(bbuff, encoding='ascii'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = SMTPEXPNCmd()
		t, bbuff = read_element(bbuff, toend= True)
		cmd.command = SMTPCommand[t.upper()]
		cmd.username = bbuff
		return cmd

	def __repr__(self):
		t = '== SMTP %s Command ==\r\n' % self.command.name
		t += 'username  : %s\r\n' % repr(self.username)
		return t


class SMTPVRFYCmd:
	def __init__(self):
		self.command   = SMTPCommand.VRFY
		self.username  = None

	def construct(self, data):
		self.username      = data

	def to_bytes(self):
		return self.command.name.encode('ascii') + b' ' +self.username.encode('ascii') +b'\r\n'

	@staticmethod
	def from_bytes(bbuff, encoding='ascii'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = SMTPVRFYCmd()
		t, bbuff = read_element(bbuff, toend= True)
		cmd.command = SMTPCommand[t.upper()]
		cmd.username = bbuff
		return cmd

	def __repr__(self):
		t = '== SMTP %s Command ==\r\n' % self.command.name
		t += 'username  : %s\r\n' % repr(self.username)
		return t



class SMTPRSETCmd:
	def __init__(self):
		self.command   = SMTPCommand.RSET

	def construct(self):
		pass

	def to_bytes(self):
		return self.command + b'\r\n'

	@staticmethod
	def from_bytes(bbuff, encoding='ascii'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = SMTPRSETCmd()
		t, bbuff = read_element(bbuff, toend= True)
		cmd.command = SMTPCommand[t.upper()]
		return cmd

	def __repr__(self):
		t = '== SMTP %s Command ==\r\n' % self.command.name
		return t


class SMTPDATACmd:
	def __init__(self):
		self.command   = SMTPCommand.DATA
		self.emaildata = None

	def construct(self, emaildata):
		self.emaildata = emaildata

	def to_bytes(self):
		return [self.command.name.encode('ascii') + b'\r\n', self.emailData.replace('\r\n','\n').encode('ascii') + b'\r\n.\r\n']

	@staticmethod
	def from_bytes(bbuff, encoding='ascii'):
		if bbuff is None:
			# we return an empty instance, as this command is muliline
			return SMTPDATACmd()
		bbuff = bbuff.decode(encoding).strip()
		cmd = SMTPDATACmd()
		t, bbuff = read_element(bbuff, marker='\n')
		cmd.command = SMTPCommand[t.strip().upper()]
		cmd.emaildata = ''
		for line in bbuff.split('\n'):
			if line.strip() == '.':
				break
			cmd.emaildata += line + '\n'
		return cmd

	def __repr__(self):
		t = '== SMTP %s Command ==\r\n' % self.command.name
		t += 'Command : %s\r\n' % self.command.name
		t += 'emaildata  : %s\r\n' % repr(self.emaildata)
		return t


class SMTPAuthStatus(enum.Enum):
	OK = enum.auto()
	NO = enum.auto()
	MORE_DATA_NEEDED = enum.auto()


class SMTPAuthMethod(enum.Enum):
	LOGIN = enum.auto()
	PLAIN = enum.auto()
	CRAM_MD5 = enum.auto()

# http://www.samlogic.net/articles/smtp-commands-reference-auth.htm
class SMTPLOGINAuth:
	def __init__(self, creds):
		self.creds = creds
		self.username = None
		self.password = None

	def update_creds(self, cmd):
		if not self.username:
			if cmd.data is not None:
				authdata = b64decode(cmd.data)
				self.username = authdata.decode()
				return SMTPAuthStatus.MORE_DATA_NEEDED, 'UGFzc3dvcmQ6'
			else:
				raise Exception('No data for username!')

		else:
			if cmd.data is not None:
				authdata = b64decode(cmd.data)
				self.password = authdata.decode()
				return self.verify_creds()

			else:
				raise Exception('No data for username!')


	def verify_creds(self):
		c = SMTPPlainCred(self.username, self.password)
		if self.creds is None:
			return SMTPAuthStatus.OK, c.toCredential()
		else:
			if c.username in self.creds:
				if self.creds[c.username] == c.password:
					return SMTPAuthStatus.OK, c.toCredential()

			else:
				return SMTPAuthStatus.NO, c.toCredential()

		return SMTPAuthStatus.NO, c.toCredential()


class SMTPPlainAuth:
	def __init__(self, creds):
		self.creds = creds
		self.username = None
		self.password = None

	def update_creds(self, cmd):
		if cmd.command == SMTPCommand.AUTH:
			if cmd.data is not None:
				authdata = b64decode(cmd.data).split(b'\x00')[1:]
				self.username = authdata[0].decode('ascii')
				self.password = authdata[1].decode('ascii')
			else:
				return SMTPAuthStatus.MORE_DATA_NEEDED, None

		else:
			authdata = b64decode(cmd.data).split(b'\x00')[1:]
			self.username = authdata[0].decode('ascii')
			self.password = authdata[1].decode('ascii')

		if self.username is not None and self.password is not None:
			return self.verify_creds()

		else:
			return SMTPAuthStatus.MORE_DATA_NEEDED, None

	def verify_creds(self):
		c = SMTPPlainCred(self.username, self.password)
		if self.creds is None:
			return SMTPAuthStatus.OK, c.toCredential()
		else:
			if c.username in self.creds:
				if self.creds[c.username] == c.password:
					return SMTPAuthStatus.OK, c.toCredential()

			else:
				return SMTPAuthStatus.NO, c.toCredential()

		return SMTPAuthStatus.NO, c.toCredential()


class SMTPPlainCred:
	def __init__(self, username, password):
		self.username = username
		self.password = password

	def toCredential(self):
		return Credential('PLAIN',
						  username=self.username,
						  password=self.password,
						  fullhash='%s:%s' % (self.username, self.password)
						  )


class SMTPAuthHandler:
	def __init__(self, authtype, creds=None, salt = None):
		if authtype == SMTPAuthMethod.PLAIN:
			self.authahndler = SMTPPlainAuth(creds)
		elif authtype == SMTPAuthMethod.CRAM_MD5:
			self.authahndler = SMTPCRAM_MD5Auth(creds, salt)
		elif authtype == SMTPAuthMethod.LOGIN:
			self.authahndler = SMTPLOGINAuth(creds)
		else:
			raise NotImplementedError

	def do_AUTH(self, cmd, salt = None):
		return self.authahndler.update_creds(cmd)

SMTPCMD = {
	SMTPCommand.HELO: SMTPHELOCmd,
	SMTPCommand.EHLO: SMTPEHLOCmd,
	SMTPCommand.MAIL: SMTPMAILCmd,
	SMTPCommand.RCPT: SMTPRCPTCmd,
	SMTPCommand.DATA: SMTPDATACmd,
	SMTPCommand.RSET: SMTPRSETCmd,
	SMTPCommand.VRFY: SMTPVRFYCmd,
	SMTPCommand.EXPN: SMTPEXPNCmd,
	SMTPCommand.HELP: SMTPHELPCmd,
	SMTPCommand.NOOP: SMTPNOOPCmd,
	SMTPCommand.QUIT: SMTPQUITCmd,
	SMTPCommand.AUTH: SMTPAUTHCmd,
	SMTPCommand.STARTTLS: SMTPSTARTTLSCmd,
	SMTPCommand.XXXX: SMTPXXXXCmd,
}